from monai.networks.nets import  BasicUNetPlusPlus

class myUNetPlusPlus:
    def __init__(self, num_classes, device):
        self.device = device
        self.num_classes = num_classes
        

    def build_model(
        self, 
        features=(16, 16, 32, 64, 128, 16),
        dropout=0.2,
        upsample='deconv',
        activation='relu'
    ):

        model = BasicUNetPlusPlus(
                spatial_dims=3,
                in_channels=1,
                out_channels=n_classes,
                deep_supervision=True,
                features=features,  # Halve the features
                dropout=dropout,
                upsample=upsample,
                act=activation
            )  
        
        return model.to(self.device)
    
    def bayesian_search(self, trial):
        
        act = trial.suggest_categorical("activation", ["LeakyReLU", "PReLU", "GELU", "ELU"])   
        dropout_prob = trial.suggest_float("dropout", 0.0, 0.5)
        upsample = trial.suggest_categorical("upsample", ["deconv", "pixelshuffle", "nontrainable"])        
        model_parameters = {"activation": act,'dropout': dropout_prob, 'upsample': upsample} 

        model = self.build_model(features, dropout_prob, upsample, act)
        return model
    
    def get_model_parameters(self):
        return {
            'model_name': 'UnetPlusPlus',
            'features': self.features,
            'dropout': self.dropout,
            'upsample': self.upsample,
            'activation': self.activation
        }


import numpy as np

class EarlyStoppingChecker:
    """
    A class to manage various early stopping criteria for model training.
    """
    
    def __init__(self, 
                 max_nan_epochs=15, 
                 plateau_patience=20, 
                 plateau_min_delta=0.001,
                 stagnation_patience=50, 
                 convergence_window=5, 
                 convergence_threshold=0.005,
                 val_interval=15,
                 monitor_metric='avg_fbeta'):
        """
        Initialize early stopping parameters.
        
        Args:
            max_nan_epochs: Maximum number of epochs with NaN loss before stopping
            plateau_patience: Number of validation checks to wait for plateau detection
            plateau_min_delta: Minimum change to qualify as improvement
            stagnation_patience: Number of validation intervals to wait for best metric improvement
            convergence_window: Window size for calculating improvement rate
            convergence_threshold: Minimum improvement rate threshold
            val_interval: Number of epochs between validation runs
            monitor_metric: Primary metric to monitor for early stopping criteria
        """
        self.max_nan_epochs = max_nan_epochs
        self.plateau_patience = plateau_patience
        self.plateau_min_delta = plateau_min_delta
        self.stagnation_patience = stagnation_patience
        self.convergence_window = convergence_window
        self.convergence_threshold = convergence_threshold
        self.val_interval = val_interval
        self.monitor_metric = monitor_metric
        
        # Counters
        self.nan_counter = 0
        
        # Flags for detailed reporting
        self.stopped_reason = None

    def check_for_nan(self, epoch_loss):
        """Check for NaN in the loss."""
        if np.isnan(epoch_loss):
            self.nan_counter += 1
            if self.nan_counter > self.max_nan_epochs:
                self.stopped_reason = f"NaN values in loss for more than {self.max_nan_epochs} epochs"
                return True
        else:
            self.nan_counter = 0  # Reset the counter if loss is valid
        return False

    def check_for_plateau(self, results):
        """Detect plateaus in validation metrics."""
        if len(results[self.monitor_metric]) < self.plateau_patience + 1:
            return False
            
        # Get the last 'patience' number of validation points
        recent_values = [x[1] for x in results[self.monitor_metric][-self.plateau_patience:]]
        # Find the max value in the window
        max_value = max(recent_values)
        # Find the min value in the window
        min_value = min(recent_values)
        
        # If the range of values is small, consider it a plateau
        if max_value - min_value < self.plateau_min_delta:
            self.stopped_reason = f"{self.monitor_metric} plateaued for {self.plateau_patience} validations"
            return True
        
        return False

    def check_best_metric_stagnation(self, results):
        """Stop if best metric hasn't improved for a number of validation intervals."""
        if "best_metric_epoch" not in results or len(results[self.monitor_metric]) < self.stagnation_patience + 1:
            return False
            
        # Get epoch of the best metric so far
        best_epoch = results["best_metric_epoch"]
        current_epoch = results[self.monitor_metric][-1][0]
        
        # Check if it's been more than 'patience' validation intervals
        if (current_epoch - best_epoch) >= (self.stagnation_patience * self.val_interval):
            self.stopped_reason = f"No improvement for {self.stagnation_patience} validation intervals"
            return True
            
        return False

    # def check_convergence_rate(self, results):
    #     """Stop when improvement rate slows below threshold."""
    #     if len(results[self.monitor_metric]) < self.convergence_window + 1:
    #         return False
        
    #     # Calculate average improvement rate over window
    #     recent_values = [x[1] for x in results[self.monitor_metric][-(self.convergence_window+1):]]
    #     improvements = [recent_values[i+1] - recent_values[i] for i in range(self.convergence_window)]
    #     avg_improvement = sum(improvements) / self.convergence_window
        
    #     if avg_improvement < self.convergence_threshold and avg_improvement > 0:
    #         self.stopped_reason = f"Convergence rate ({avg_improvement:.6f}) below threshold"
    #         return True
            
    #     return False

    def should_stop_training(self, epoch_loss, results=None, check_metrics=False):
        """
        Comprehensive check for whether training should stop.
        
        Args:
            epoch_loss: Current epoch's loss value
            results: Dictionary containing training metrics history
            check_metrics: Whether to also check validation metrics-based criteria
            
        Returns:
            bool: True if training should stop, False otherwise
        """
        # Check for NaN in loss (can be done every epoch)
        if self.check_for_nan(epoch_loss):
            return True
            
        # Only check metric-based criteria if requested and results are provided
        if check_metrics and results:
            # Check for plateau in validation metrics
            if self.check_for_plateau(results):
                return True
                
            # Check if best metric hasn't improved for a while
            if self.check_best_metric_stagnation(results):
                return True
                
            # # Check if convergence rate has slowed down
            # if self.check_convergence_rate(results):
            #     return True
                
        return False
    
    def get_stopped_reason(self):
        """Get the reason for stopping, if any."""
        if self.stopped_reason:
            return f"Early stopping triggered: {self.stopped_reason}"
        return "No early stopping criteria met."


import torch.nn as nn
import torch

class myModelTemplate:
    def __init__(self):
        """
        Initialize the model template.
        """
        # Placeholder for the model and config
        self.model = None
        self.config = None

    def build_model(self, config: dict):
        """
        Build the model based on provided parameters in a config dictionary.
        """
        pass

    def bayesian_search(self, trial, num_classes: int):
        """
        Define the hyperparameter search space for Bayesian optimization and build the model.

        The search space below is just an example and can be customized.
        
        Args:
            trial (optuna.trial.Trial): Optuna trial object.
            num_classes (int): Number of classes in the dataset.
        """
        pass
    
    def get_model_parameters(self):
        """Retrieve stored model parameters."""
        if self.model is None:
            raise ValueError("Model has not been initialized yet. Call build_model() or bayesian_search() first.")
        
        return self.config



import os
import argparse
import copick
import torch
from tqdm import tqdm
from typing import Optional, Union, Tuple, List
from collections import defaultdict
import pytorch_lightning as pl
import torch.distributed as dist
from pytorch_lightning import Trainer
from pytorch_lightning.loggers import MLFlowLogger
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor
from monai.data import DataLoader, Dataset, CacheDataset, decollate_batch
from dotenv import load_dotenv
from monai.transforms import (
    Compose,
    EnsureChannelFirstd,
    Orientationd,
    AsDiscrete,
    RandFlipd,
    RandRotate90d,
    NormalizeIntensityd,
    NormalizeIntensityd,
    RandCropByLabelClassesd,    
)
from monai.networks.nets import UNet
from monai.losses import TverskyLoss
from monai.metrics import DiceMetric, ConfusionMatrixMetric
import optuna
from optuna.integration import PyTorchLightningPruningCallback


def get_args():
    parser = argparse.ArgumentParser(
        description = "Hyperparamter tuning using PyTorch Lightning distributed data-parallel and Optuna."
    )
    parser.add_argument('--copick_config_path', type=str, default='copick_config_dataportal_10439.json')
    parser.add_argument('--copick_user_name', type=str, default='user0')
    parser.add_argument('--copick_segmentation_name', type=str, default='paintedPicks')
    parser.add_argument('--train_batch_size', type=int, default=1)
    parser.add_argument('--val_batch_size', type=int, default=1)
    parser.add_argument('--num_random_samples_per_batch', type=int, default=16)
    parser.add_argument('--learning_rate', type=float, default=1e-4)
    parser.add_argument('--num_epochs', type=int, default=20)
    parser.add_argument('--num_gpus', type=int, default=1)
    parser.add_argument('--num_optuna_trials', type=int, default=10)
    parser.add_argument('--pruning', action="store_true", help="Activate the pruning feature. `MedianPruner` stops unpromising trials at the early stages of training.")
    return parser.parse_args()


class Model(pl.LightningModule):
    def __init__(
        self, 
        spatial_dims: int = 3,
        in_channels: int = 1,
        out_channels: int = 8,
        channels: Union[Tuple[int, ...], List[int]] = (48, 64, 80, 80),
        strides: Union[Tuple[int, ...], List[int]] = (2, 2, 1),
        num_res_units: int = 1,
        lr: float=1e-3):
    
        super().__init__()
        self.save_hyperparameters()
        self.model = UNet(
            spatial_dims=self.hparams.spatial_dims,
            in_channels=self.hparams.in_channels,
            out_channels=self.hparams.out_channels,
            channels=self.hparams.channels,
            strides=self.hparams.strides,
            num_res_units=self.hparams.num_res_units,
        )
        self.loss_fn = TverskyLoss(include_background=True, to_onehot_y=True, softmax=True)  # softmax=True for multiclass
        self.metric_fn = DiceMetric(include_background=False, reduction="mean", ignore_empty=True)

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch['image'], batch['label']
        y_hat = self(x)
        loss = self.loss_fn(y_hat, y)
        return loss
    
    def validation_step(self, batch, batch_idx):
        with torch.no_grad(): # This ensures that gradients are not stored in memory
            x, y = batch['image'], batch['label']
            y_hat = self(x)
            metric_val_outputs = [AsDiscrete(argmax=True, to_onehot=self.hparams.out_channels)(i) for i in decollate_batch(y_hat)]
            metric_val_labels = [AsDiscrete(to_onehot=self.hparams.out_channels)(i) for i in decollate_batch(y)]

            # compute metric for current iteration
            self.metric_fn(y_pred=metric_val_outputs, y=metric_val_labels)
            metrics = self.metric_fn.aggregate(reduction="mean_batch")
            for i,m in enumerate(metrics):
                self.log(f"validation metric class {i+1}", m, prog_bar=True, on_epoch=True, sync_dist=True)
            metric = torch.mean(metrics) # cannot log ndarray 
            self.log('val_metric', metric, prog_bar=True, on_epoch=True, sync_dist=True) # sync_dist=True for distributed training
        return {'val_metric': metric}

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.hparams.lr)
    

class CopickDataModule(pl.LightningDataModule):
    def __init__(
        self, 
        copick_config_path: str, 
        train_batch_size: int,
        val_batch_size: int,
        num_random_samples_per_batch: int):
    
        super().__init__()
        self.train_batch_size = train_batch_size
        self.val_batch_size = val_batch_size
        
        self.data_dicts, self.nclasses = self.data_from_copick(copick_config_path)
        self.train_files = self.data_dicts[:int(len(self.data_dicts)//2)]
        self.val_files = self.data_dicts[int(len(self.data_dicts)//2):]
        print(f"Number of training samples: {len(self.train_files)}")
        print(f"Number of validation samples: {len(self.val_files)}")

        # Non-random transforms to be cached
        self.non_random_transforms = Compose([
            EnsureChannelFirstd(keys=["image", "label"], channel_dim="no_channel"),
            NormalizeIntensityd(keys="image"),
            Orientationd(keys=["image", "label"], axcodes="RAS")
        ])

        # Random transforms to be applied during training
        self.random_transforms = Compose([
            RandCropByLabelClassesd(
                keys=["image", "label"],
                label_key="label",
                spatial_size=[96, 96, 96],
                num_classes=self.nclasses,
                num_samples=num_random_samples_per_batch 
            ),
            RandRotate90d(keys=["image", "label"], prob=0.5, spatial_axes=[0, 2]),
            RandFlipd(keys=["image", "label"], prob=0.5, spatial_axis=0),    
        ])

    def setup(self, stage: Optional[str] = None) -> None:
        self.train_ds = CacheDataset(data=self.train_files, transform=self.non_random_transforms, cache_rate=1.0)
        self.train_ds = Dataset(data=self.train_ds, transform=self.random_transforms)
        self.val_ds = CacheDataset(data=self.val_files, transform=self.non_random_transforms, cache_rate=1.0)
        self.val_ds = Dataset(data=self.val_ds, transform=self.random_transforms)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
                self.train_ds,
                batch_size=self.train_batch_size,
                shuffle=True,
                num_workers=4,
                persistent_workers=True,
                pin_memory=torch.cuda.is_available(),
            )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
                self.val_ds,
                batch_size=self.val_batch_size,
                shuffle=False,  # Ensure the data order remains consistent
                num_workers=4,
                persistent_workers=True,
                pin_memory=torch.cuda.is_available(),
            )

    @staticmethod
    def data_from_copick(copick_config_path):
        root = copick.from_file(copick_config_path)
        nclasses = len(root.pickable_objects) + 1
        data_dicts = []
        target_objects = defaultdict(dict)
        for object in root.pickable_objects:
            if object.is_particle:
                target_objects[object.name]['label'] = object.label
                target_objects[object.name]['radius'] = object.radius
        
        data_dicts = []
        for run in tqdm(root.runs[:8]):
            tomogram = run.get_voxel_spacing(10).get_tomogram('wbp').numpy()
            segmentation = run.get_segmentations(name='paintedPicks', user_id='user0', voxel_size=10, is_multilabel=True)[0].numpy()
            membrane_seg = run.get_segmentations(name='membrane', user_id="data-portal")[0].numpy()
            segmentation[membrane_seg==1] = 1  
            data_dicts.append({"image": tomogram, "label": segmentation})
        
        return data_dicts, nclasses


def objective(trial: optuna.trial.Trial) -> float:
    args = get_args()
    mlf_logger = MLFlowLogger(experiment_name='training-3D-UNet-model-for-the-cryoET-ML-Challenge',
                              tracking_uri='http://mlflow.mlflow.svc.cluster.local:5000',
                              #run_name='test1'
                              )

    # Trainer callbacks
    checkpoint_callback = ModelCheckpoint(monitor='val_metric', save_top_k=1, mode='max')
    lr_monitor = LearningRateMonitor(logging_interval='epoch')

    # Detect distributed training environment
    devices = list(range(args.num_gpus))

    #channels, strides_pattern, num_res_units = sync_hyperparameters(trial)
    # We optimize the number of layers, strides, and number of residual units
    num_layers = trial.suggest_int("num_layers", 3, 5)
    base_channel = trial.suggest_categorical("base_channel", [8, 16, 32, 64])
    channels = [base_channel * (2 ** i) for i in range(num_layers)]
    num_downsampling_layers = trial.suggest_int("num_downsampling_layers", 1, num_layers - 1)
    strides_pattern = [2] * num_downsampling_layers + [1] * (num_layers - num_downsampling_layers - 1)        
    num_res_units = trial.suggest_int("num_res_units", 1, 3)
    
    model = Model(channels=channels, strides=strides_pattern, num_res_units=num_res_units, lr=args.learning_rate)
    datamodule = CopickDataModule(args.copick_config_path, args.train_batch_size, args.val_batch_size, args.num_random_samples_per_batch)
    callback = PyTorchLightningPruningCallback(trial, monitor="val_metric")

    # Priotize performace over precision
    torch.set_float32_matmul_precision('medium') # or torch.set_float32_matmul_precision('high')    
    
    # Trainer for distributed training with DDP
    trainer = Trainer(
        max_epochs=args.num_epochs,
        logger=mlf_logger,
        callbacks=[checkpoint_callback, lr_monitor],
        strategy="ddp_spawn",
        accelerator="gpu",
        devices=devices,
        num_nodes=1, #int(os.environ.get("WORLD_SIZE", 1)) // args.num_gpus,
        log_every_n_steps=1
    )

    hyperparameters = dict(op_num_layers=num_layers, op_base_channel=base_channel, op_num_downsampling_layers=num_downsampling_layers, op_num_res_units=num_res_units)
    trainer.logger.log_hyperparams(hyperparameters)
    trainer.fit(model, datamodule=datamodule)
    callback.check_pruned()
    return trainer.callback_metrics["val_metric"].item()


if __name__ == "__main__":
    args = get_args()
    # MLflow setup
    username = os.getenv('MLFLOW_TRACKING_USERNAME')
    password = os.getenv('MLFLOW_TRACKING_PASSWORD')
    if not password or not username:
        print("Password not found in environment, loading from .env file...")
        load_dotenv()  # Loads environment variables from a .env file
        username = os.getenv('MLFLOW_TRACKING_USERNAME')
        password = os.getenv('MLFLOW_TRACKING_PASSWORD')
        
    # Check again after loading .env file
    if not password:
        raise ValueError("Password is not set in environment variables or .env file!")
    else:
        print("Password loaded successfully")
        os.environ['MLFLOW_TRACKING_USERNAME'] = username
        os.environ['MLFLOW_TRACKING_PASSWORD'] = password    
    
    pruner: optuna.pruners.BasePruner = (
        optuna.pruners.MedianPruner() if args.pruning else optuna.pruners.NopPruner()
    )
    storage = "sqlite:///example.db"
    study = optuna.create_study(
        study_name="pl_ddp",
        storage=storage,
        direction="maximize",
        load_if_exists=True,
        pruner=pruner
    )
    study.optimize(objective, n_trials=args.num_optuna_trials)

    # Print the best hyperparameters
    print(f"Best trial: {study.best_trial.value}")
    print(f"Best hyperparameters: {study.best_trial.params}")


from octopi.entry_points import run_train, run_segment_predict, run_localize, run_optuna
from octopi.submit_slurm import create_shellsubmit, create_multiconfig_shellsubmit
from octopi.processing.importers import cli_mrcs_parser, cli_dataportal_parser
from octopi.entry_points import common 
from octopi import utils
import argparse

def create_train_script(args):
    """
    Create a SLURM script for training 3D CNN models
    """

    strconfigs = ""
    for config in args.config:
        strconfigs += f"--config {config}"

    command = f"""
octopi train \\
    {strconfigs} \\
    --model-save-path {args.model_save_path} \\
    --target-info {','.join(args.target_info)} \\
    --voxel-size {args.voxel_size} --tomo-alg {args.tomo_alg} --Nclass {args.Nclass} \\
    --tomo-batch-size {args.tomo_batch_size} --num-tomo-crops {args.num_tomo_crops} \\
    --best-metric {args.best_metric} --num-epochs {args.num_epochs} --val-interval {args.val_interval} \\
    """

    # If a model config is provided, use it to build the model
    if args.model_config is not None:
        command += f" --model-config {args.model_config}"
    else:
        channels = ",".join(map(str, args.channels))
        strides = ",".join(map(str, args.strides))
        command += (
            f" --tversky-alpha {args.tversky_alpha}"
            f" --channels {channels}"
            f" --strides {strides}"
            f" --dim-in {args.dim_in}"
            f" --res-units {args.res_units}"
        )

    # If Model Weights are provided, use them to initialize the model
    if args.model_weights is not None and args.model_config is not None:
        command += f" --model-weights {args.model_weights}"

    create_shellsubmit(
        job_name = args.job_name,
        output_file = 'trainer.log',
        shell_name = 'train_octopi.sh',
        conda_path = args.conda_env,
        command = command,
        num_gpus = 1,
        gpu_constraint = args.gpu_constraint
    )

def train_model_slurm():
    """
    Create a SLURM script for training 3D CNN models
    """
    parser_description = "Create a SLURM script for training 3D CNN models"
    args = run_train.train_model_parser(parser_description, add_slurm=True)
    create_train_script(args)   

def create_model_explore_script(args):
    """
    Create a SLURM script for running bayesian optimization on 3D CNN models
    """
    strconfigs = ""
    for config in args.config:
        strconfigs += f"--config {config}"

    command = f"""
octopi model-explore \\
    --model-type {args.model_type} --model-save-path {args.model_save_path} \\
    --voxel-size {args.voxel_size} --tomo-alg {args.tomo_alg} --Nclass {args.Nclass} \\
    --val-interval {args.val_interval} --num-epochs {args.num_epochs} --num-trials {args.num_trials} \\
    --best-metric {args.best_metric} --mlflow-experiment-name {args.mlflow_experiment_name} \\
    --target-name {args.target_name} --target-session-id {args.target_session_id} --target-user-id {args.target_user_id} \\
    {strconfigs}
"""

    create_shellsubmit(
        job_name = args.job_name,
        output_file = 'optuna.log',
        shell_name = 'model_explore.sh',
        conda_path = args.conda_env,
        command = command,
        num_gpus = 1,
        gpu_constraint = args.gpu_constraint
    )

def model_explore_slurm():
    """
    Create a SLURM script for running bayesian optimization on 3D CNN models
    """
    parser_description = "Create a SLURM script for running bayesian optimization on 3D CNN models"
    args = run_optuna.optuna_parser(parser_description, add_slurm=True)
    create_model_explore_script(args)   

def create_inference_script(args):
    """
    Create a SLURM script for running inference on 3D CNN models
    """

    if len(args.config.split(',')) > 1:

        create_multiconfig_shellsubmit(
            job_name = args.job_name,
            output_file = 'predict.log',
            shell_name = 'segment.sh',
            conda_path = args.conda_env,
            base_inputs = args.base_inputs,
            config_inputs = args.config_inputs,
            command = args.command,
            num_gpus = args.num_gpus,
            gpu_constraint = args.gpu_constraint
        )
    else:

        command = f"""
octopi inference \\
    --config {args.config} \\
    --seg-info  {",".join(args.seg_info)} \\
    --model-weights {args.model_weights} \\
    --dim-in {args.dim_in} --res-units {args.res_units} \\
    --model-type {args.model_type} --channels {",".join(map(str, args.channels))} --strides {",".join(map(str, args.strides))} \\
    --voxel-size {args.voxel_size} --tomo-alg {args.tomo_alg} --Nclass {args.Nclass}
"""

        create_shellsubmit(
            job_name = args.job_name,
            output_file = 'predict.log',
            shell_name = 'segment.sh',
            conda_path = args.conda_env,
            command = command,
            num_gpus = 1,
            gpu_constraint = args.gpu_constraint
        )

def inference_slurm():
    """
    Create a SLURM script for running segmentation predictions with a specified model and configuration on CryoET Tomograms.
    """
    parser_description = "Create a SLURM script for running segmentation predictions with a specified model and configuration on CryoET Tomograms."
    args = run_segment_predict.inference_parser(parser_description, add_slurm=True)
    create_inference_script(args)

def create_localize_script(args):
    """"
    Create a SLURM script for running localization on predicted segmentation masks
    """
    if len(args.config.split(',')) > 1:
    
        create_multiconfig_shellsubmit(
            job_name = args.job_name,
            output_file = args.output,
            shell_name = args.output_script,
            conda_path = args.conda_env,
            base_inputs = args.base_inputs,
            config_inputs = args.config_inputs,
            command = args.command
        )
    else:

        command = f"""
octopi localize \\
    --config {args.config} \\
    --voxel-size {args.voxel_size} --pick-session-id {args.pick_session_id} --pick-user-id {args.pick_user_id} \\
    --method {args.method}  --seg-info {",".join(args.seg_info)} \\
"""
        if args.pick_objects is not None:
            command += f" --pick-objects {args.pick_objects}"

        create_shellsubmit(
            job_name = args.job_name,
            output_file = 'localize.log',
            shell_name = 'localize.sh',
            conda_path = args.conda_env,
            command = command,
            num_gpus = 0
        )

def localize_slurm():
    """
    Create a SLURM script for running localization on predicted segmentation masks
    """
    parser_description = "Create a SLURM script for localization on predicted segmentation masks"
    args = run_localize.localize_parser(parser_description, add_slurm=True)
    create_localize_script(args)
        
def create_extract_mb_picks_script(args):
    pass

def extract_mb_picks_slurm():
    pass


def create_import_mrc_script(args):
    """
    Create a SLURM script for importing mrc volumes and potentialy downsampling
    """
    command = f"""
octopi import-mrc-volumes \\
    --mrcs-path {args.mrcs_path} \\
    --config {args.config} --target-tomo-type {args.target_tomo_type} \\
    --input-voxel-size {args.input_voxel_size} --output-voxel-size {args.output_voxel_size}
"""

    create_shellsubmit(
        job_name = args.job_name,
        output_file = 'importer.log',
        shell_name = 'mrc_importer.sh',
        conda_path = args.conda_env,
        command = command
    )

def import_mrc_slurm():
    """
    Create a SLURM script for importing mrc volumes and potentialy downsampling
    """
    parser_description = "Create a SLURM script for importing mrc volumes and potentialy downsampling"
    args = cli_mrcs_parser(parser_description, add_slurm=True)
    create_import_mrc_script(args)


def create_download_dataportal_script(args):
    """
    Create a SLURM script for downloading tomograms from the Dataportal
    """
    command = f"""
octopi download-dataportal \\
    --config {args.config} --datasetID {args.datasetID} \\
    --overlay-path {args.overlay_path} 
    --dataportal-name {args.dataportal_name} --target-tomo-type {args.target_tomo_type} \\
    --input-voxel-size {args.input_voxel_size} --output-voxel-size {args.output_voxel_size}
"""

    create_shellsubmit(
        job_name = args.job_name,
        output_file = 'importer.log',
        shell_name = 'dataportal_importer.sh',
        conda_path = args.conda_env,
        command = command
    )

def download_dataportal_slurm():
    """
    Create a SLURM script for downloading tomograms from the Dataportal
    """
    parser_description = "Create a SLURM script for downloading tomograms from the Dataportal"
    args = cli_dataportal_parser(parser_description, add_slurm=True)
    create_download_dataportal_script(args)



from octopi.datasets import dataset, augment, cached_datset
from octopi.datasets.generators import TrainLoaderManager
from monai.data import DataLoader, SmartCacheDataset, CacheDataset, Dataset
from octopi import io
import multiprocess as mp
from typing import List
from tqdm import tqdm
import torch, gc

class MultiConfigTrainLoaderManager(TrainLoaderManager):

    def __init__(self, 
                 configs: dict,  # Dictionary of session names and config paths
                 target_name: str,
                 target_session_id: str = None,
                 target_user_id: str = None,
                 voxel_size: float = 10, 
                 tomo_algorithm: List[str] = ['wbp'], 
                 tomo_batch_size: int = 15, 
                 Nclasses: int = 3):
        """
        Initialize MultiConfigTrainLoaderManager with multiple configs.

        Args:
            configs (list): List of config file paths.
            Other arguments are inherited from TrainLoaderManager.
        """

        # Initialize shared attributes manually (skip super().__init__ to avoid invalid config handling)
        self.config = configs
        self.roots = {name: io.load_copick_config(path) for name, path in configs.items()}

        # Target and algorithm parameters
        self.target_name = target_name
        self.target_session_id = target_session_id
        self.target_user_id = target_user_id
        self.voxel_size = voxel_size
        self.tomo_algorithm = tomo_algorithm

        # Data management parameters
        self.Nclasses = Nclasses
        self.tomo_batch_size = tomo_batch_size
        self.reload_training_dataset = True
        self.reload_validation_dataset = True
        self.val_loader = None
        self.train_loader = None

        # Initialize Run IDs placeholder
        self.myRunIDs = {}

        # Initialize the input dimensions   
        self.nx = None
        self.ny = None
        self.nz = None        

    def get_available_runIDs(self):
        """
        Identify and return a combined list of run IDs with available segmentations 
        across all configured CoPick projects.

        Returns:
            List of tuples: [(session_name, run_name), ...]
        """
        available_runIDs = []  
        for name, root in self.roots.items():
            runIDs = [run.name for run in root.runs]                       
            for run in runIDs:
                run = root.get_run(run)
                seg = run.get_segmentations(
                    name=self.target_name, 
                    session_id=self.target_session_id, 
                    user_id=self.target_user_id,
                    voxel_size=float(self.voxel_size)
                )
                if len(seg) > 0:
                    available_runIDs.append((name, run.name))  # Include session name for disambiguation

        # If No Segmentations are Found, Inform the User
        if len(available_runIDs) == 0:
            print(
                f"[Error] No segmentations found for the target query:\n"
                f"TargetName: {self.target_name}, UserID: {self.target_user_id}, "
                f"SessionID: {self.target_session_id}\n"
                f"Please check the target name, user ID, and session ID.\n"
            )
            exit()

        return available_runIDs        

    def get_data_splits(self, 
                        trainRunIDs: str = None,
                        validateRunIDs: str = None,
                        train_ratio: float = 0.8,
                        val_ratio: float = 0.1,
                        test_ratio: float = 0.1,
                        create_test_dataset: bool = True):
        """
        Override to handle run IDs as (session_name, run_name) tuples.
        """
        # Use the get_available_runIDs method to handle multiple projects
        runIDs = self.get_available_runIDs()
        return super().get_data_splits(trainRunIDs = runIDs, 
                                       train_ratio = train_ratio, 
                                       val_ratio = val_ratio, 
                                       test_ratio = test_ratio, 
                                       create_test_dataset = create_test_dataset)

    def _initialize_train_iterators(self):
        """
        Initialize the training data iterators with multi-config support.
        """
        self.padded_train_list = self._get_padded_list(self.myRunIDs['train'], self.train_batch_size)
        self.train_data_iter = iter(self._get_data_batches(self.padded_train_list, self.train_batch_size))

    def _initialize_val_iterators(self):
        """
        Initialize the validation data iterators with multi-config support.
        """
        self.padded_val_list = self._get_padded_list(self.myRunIDs['validate'], self.val_batch_size)
        self.val_data_iter = iter(self._get_data_batches(self.padded_val_list, self.val_batch_size))

    def _load_data(self, runIDs):
        """
        Load data from multiple CoPick projects for given run IDs.

        Args:
            runIDs (list): List of (session_name, run_name) tuples.

        Returns:
            List: Combined dataset for the specified run IDs.
        """

        data = []
        for session_name, run_name in tqdm(runIDs):
            root = self.roots[session_name]
            data.extend(io.load_training_data(
                root, [run_name], self.voxel_size, self.tomo_algorithm, 
                self.target_name, self.target_session_id, self.target_user_id, 
                progress_update=False ))
        self._check_max_label_value(data)
        return data

    def create_train_dataloaders(self, *args, **kwargs):
        """
        Override data loading to fetch from multiple projects.
        """
        my_crop_size = kwargs.get("crop_size", 96)
        my_num_samples = kwargs.get("num_samples", 128)

        # If reloads are disabled and loaders already exist, reuse them
        if self.reload_frequency < 0 and (self.train_loader is not None) and (self.val_loader is not None):
            return self.train_loader, self.val_loader        

        # Estimate Max Number of Threads with mp.cpu_count
        n_procs = min(mp.cpu_count(), 4)               

        if self.train_loader is None: 
            # Fetch the next batch of run IDs
            trainRunIDs = self._extract_run_ids('train_data_iter', self._initialize_train_iterators)
            train_files = self._load_data(trainRunIDs)

            # # Create the cached dataset with non-random transforms
            train_ds = SmartCacheDataset(data=train_files, transform=augment.get_transforms(), cache_rate=0.5)

            # # Delete the training files to free memory
            train_files = None
            gc.collect()

            # Create the cached dataset with non-random transforms
            # train_ds = cached_datset.MultiConfigCacheDataset(
            #     self, trainRunIDs, transform=augment.get_transforms(), cache_rate=1.0
            # )

            # I need to read (nx,ny,nz) and scale the crop size to make sure it isnt larger than nx.
            if self.nx is None: (self.nx,self.ny,self.nz) = train_ds[0]['image'].shape[1:]
            self.input_dim = io.get_input_dimensions(train_ds, my_crop_size)

            # Wrap the cached dataset to apply random transforms during iteration
            self.dynamic_train_dataset = dataset.DynamicDataset(
                data=train_ds, 
                transform=augment.get_random_transforms(self.input_dim, my_num_samples, self.Nclasses)
            ) 

            self.train_loader = DataLoader(
                self.dynamic_train_dataset,
                batch_size=1,
                shuffle=True,
                num_workers=n_procs,
                pin_memory=torch.cuda.is_available(),             
            )

        else:
            # Fetch the next batch of run IDs
            trainRunIDs = self._extract_run_ids('train_data_iter', self._initialize_train_iterators)
            train_files = self._load_data(trainRunIDs)
            train_ds = CacheDataset(data=train_files, transform=augment.get_transforms(), cache_rate=1.0)
            self.dynamic_train_dataset.update_data(train_ds)

        # We Only Need to Reload the Validation Dataset if the Total Number of Runs is larger than 
        # the tomo batch size
        if self.val_loader is None: 

            validateRunIDs = self._extract_run_ids('val_data_iter', self._initialize_val_iterators)             
            val_files  = self._load_data(validateRunIDs)    

            # # Create validation dataset
            val_ds = SmartCacheDataset(data=val_files, transform=augment.get_transforms(), cache_rate=1.0)            

            # # Delete the validation files to free memory
            val_files = None
            gc.collect()

            # Create the cached dataset with non-random transforms
            # val_ds = cached_datset.MultiConfigCacheDataset(
            #     self, validateRunIDs, transform=augment.get_transforms(), cache_rate=1.0
            # )

            # # I need to read (nx,ny,nz) and scale the crop size to make sure it isnt larger than nx.
            # if self.nx is None:
            #     (self.nx,self.ny,self.nz) = val_ds[0]['image'].shape[1:]

            # if crop_size > self.nx: self.input_dim = (self.nx, crop_size, crop_size)
            # else:                   self.input_dim = (crop_size, crop_size, crop_size)

            # Wrap the cached dataset to apply random transforms during iteration
            self.dynamic_validation_dataset = dataset.DynamicDataset( data=val_ds )

            # Create validation DataLoader
            self.val_loader  = DataLoader(
                self.dynamic_validation_dataset,
                batch_size=1,
                num_workers=n_procs,
                pin_memory=torch.cuda.is_available(),
                shuffle=False,  # Ensure the data order remains consistent            
            )
        else:
            validateRunIDs = self._extract_run_ids('val_data_iter', self._initialize_val_iterators)             
            val_files   = self._load_data(validateRunIDs)

            val_ds = CacheDataset(data=val_files, transform=augment.get_transforms(), cache_rate=1.0)
            self.dynamic_validation_dataset.update_data(val_ds)
            
        return self.train_loader, self.val_loader


    def tmp_return_datasets(self):
        trainRunIDs = self._extract_run_ids('train_data_iter', self._initialize_train_iterators)
        train_files = self._load_data(trainRunIDs)   

        validateRunIDs = self._extract_run_ids('val_data_iter', self._initialize_val_iterators)
        val_files = self._load_data(validateRunIDs)

        return train_files, val_files        


from octopi.datasets import generators, multi_config_generator
from octopi.pytorch import hyper_search
import torch, mlflow, optuna
from octopi import utils
from typing import List
import pandas as pd

class ModelSearchSubmit:
    def __init__(
        self,
        copick_config: str,
        target_name: str,
        target_user_id: str,
        target_session_id: str,
        tomo_algorithm: str,
        voxel_size: float,
        Nclass: int,
        model_type: str,
        best_metric: str = 'avg_f1',
        num_epochs: int = 1000,
        num_trials: int = 100,
        data_split: str = 0.8,
        random_seed: int = 42,
        val_interval: int = 10,
        tomo_batch_size: int = 15,
        trainRunIDs: List[str] = None,
        validateRunIDs: List[str] = None,
        mlflow_experiment_name: str = 'explore',
    ):
        """
        Initialize the ModelSearch class for architecture search with Optuna.

        Args:
            copick_config (str or dict): Path to the CoPick configuration file or a dictionary for multi-config training.
            target_name (str): Name of the target for segmentation.
            target_user_id (str): Optional user ID for tracking.
            target_session_id (str): Optional session ID for tracking.
            tomo_algorithm (str): Tomogram algorithm to use.
            voxel_size (float): Voxel size for tomograms.
            Nclass (int): Number of prediction classes.
            model_type (str): Type of model to use.
            mlflow_experiment_name (str): MLflow experiment name.
            random_seed (int): Seed for reproducibility.
            num_epochs (int): Number of epochs per trial.
            num_trials (int): Number of trials for hyperparameter optimization.
            tomo_batch_size (int): Batch size for tomogram loading.
            best_metric (str): Metric to optimize.
            val_interval (int): Validation interval.
            trainRunIDs (List[str]): List of training run IDs.
            validateRunIDs (List[str]): List of validation run IDs.
            data_split (str): Data split ratios.
        """

        # Input parameters 
        self.copick_config = copick_config
        self.target_name = target_name
        self.target_user_id = target_user_id
        self.target_session_id = target_session_id
        self.tomo_algorithm = tomo_algorithm
        self.voxel_size = voxel_size
        self.Nclass = Nclass
        self.model_type = model_type
        self.mlflow_experiment_name = mlflow_experiment_name
        self.random_seed = random_seed
        self.num_epochs = num_epochs
        self.num_trials = num_trials
        self.tomo_batch_size = tomo_batch_size
        self.best_metric = best_metric
        self.val_interval = val_interval
        self.trainRunIDs = trainRunIDs
        self.validateRunIDs = validateRunIDs
        self.data_split = data_split
        
        # Data generator - will be initialized in _initialize_data_generator()
        self.data_generator = None

        # Set random seed for reproducibility
        utils.set_seed(self.random_seed)

        # Initialize dataset generator
        self._initialize_data_generator()

    def _initialize_data_generator(self):
        """Initializes the data generator for training and validation datasets."""
        self._print_input_configs()

        if isinstance(self.copick_config, dict):
            self.data_generator = multi_config_generator.MultiConfigTrainLoaderManager(
                self.copick_config,
                self.target_name,
                target_session_id=self.target_session_id,
                target_user_id=self.target_user_id,
                tomo_algorithm=self.tomo_algorithm,
                voxel_size=self.voxel_size,
                Nclasses=self.Nclass,
                tomo_batch_size=self.tomo_batch_size
            )
        else:
            self.data_generator = generators.TrainLoaderManager(
                self.copick_config,
                self.target_name,
                target_session_id=self.target_session_id,
                target_user_id=self.target_user_id,
                tomo_algorithm=self.tomo_algorithm,
                voxel_size=self.voxel_size,
                Nclasses=self.Nclass,
                tomo_batch_size=self.tomo_batch_size
            )

        # Split datasets into training and validation
        ratios = utils.parse_data_split(self.data_split)
        self.data_generator.get_data_splits(
            trainRunIDs=self.trainRunIDs,
            validateRunIDs=self.validateRunIDs,
            train_ratio = ratios[0], val_ratio = ratios[1], test_ratio = ratios[2],
            create_test_dataset = False
        )
        
        # Get the reload frequency
        self.data_generator.get_reload_frequency(self.num_epochs)
        
    def _print_input_configs(self):
        """Prints training configuration for debugging purposes."""
        print(f'\nTraining with:')
        if isinstance(self.copick_config, dict):
            for session, config in self.copick_config.items():
                print(f'  {session}: {config}')
        else:
            print(f'  {self.copick_config}')
        print()

    def run_model_search(self):
        """Performs model architecture search using Optuna and MLflow."""

        # Set up MLflow tracking
        try:
            tracking_uri = utils.mlflow_setup()
            mlflow.set_tracking_uri(tracking_uri)
        except Exception as e:
            print(f'Failed to set up MLflow tracking: {e}')
            pass

        mlflow.set_experiment(self.mlflow_experiment_name)
    
        # Create a storage object with heartbeat configuration
        storage_url = f"sqlite:///explore_results_{self.model_type}/trials.db"
        self.storage = optuna.storages.RDBStorage(
            url=storage_url,
            heartbeat_interval=60,  # Record heartbeat every minute
            grace_period=600,       # 10 minutes grace period
            failed_trial_callback=optuna.storages.RetryFailedTrialCallback(max_retry=1)
        )

        # Detect GPU availability
        gpu_count = torch.cuda.device_count()
        print(f'Running Architecture Search Over {gpu_count} GPUs\n')

        # Initialize model search object
        if gpu_count > 1:
            self._multi_gpu_optuna(gpu_count)
        else:
            model_search = hyper_search.BayesianModelSearch(self.data_generator, self.model_type)
            self._single_gpu_optuna(model_search)

    def _single_gpu_optuna(self, model_search):
        """Runs Optuna optimization on a single GPU."""
        
        with mlflow.start_run(nested=False) as parent_run:

            model_search.parent_run_id = parent_run.info.run_id
            model_search.parent_run_name = parent_run.info.run_name

            # Log the experiment parameters
            mlflow.log_params({"random_seed": self.random_seed})
            mlflow.log_params(self.data_generator.get_dataloader_parameters())
            mlflow.log_params({"parent_run_name": parent_run.info.run_name})

            # Determine device
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

            # Create the study and run the optimization
            study = self.get_optuna_study()
            study.optimize(
                lambda trial: model_search.objective(
                    trial, self.num_epochs, device,
                    val_interval=self.val_interval,
                    best_metric=self.best_metric,
                ),
                n_trials=self.num_trials
            )

            # Save contour plot
            self.save_contour_plot_as_png(study)

            print(f"Best trial: {study.best_trial.value}")
            print(f"Best params: {study.best_params}")

    def _multi_gpu_optuna(self, gpu_count):
        """Runs Optuna optimization on multiple GPUs."""
        with mlflow.start_run() as parent_run:

            # Log the experiment parameters
            mlflow.log_params({"random_seed": self.random_seed})
            mlflow.log_params(self.data_generator.get_dataloader_parameters())
            mlflow.log_params({"parent_run_name": parent_run.info.run_name})

            # Run multi-GPU optimization
            study = self.get_optuna_study()
            study.optimize(
                lambda trial: hyper_search.BayesianModelSearch(self.data_generator, self.model_type).multi_gpu_objective(
                    parent_run, trial,
                    self.num_epochs,
                    best_metric=self.best_metric,
                    val_interval=self.val_interval,
                    gpu_count=gpu_count
                ),
                n_trials=self.num_trials,
                n_jobs=gpu_count
            )

            # Save contour Plot
            self.save_contour_plot_as_png(study)

        print(f"Best trial: {study.best_trial.value}")
        print(f"Best params: {study.best_params}")

    def _get_optuna_sampler(self):
        """Returns Optuna's TPE sampler with default settings."""
        return optuna.samplers.TPESampler(
            n_startup_trials=10,
            n_ei_candidates=24,
            multivariate=True
        )
        # return optuna.samplers.BoTorchSampler(
        #     n_startup_trials=10,
        #     multivariate=True
        # )

    def get_optuna_study(self):
        """Returns the Optuna study object."""
        return optuna.create_study(
                storage=self.storage,
                direction="maximize",
                sampler=self._get_optuna_sampler(),
                load_if_exists=True,
                pruner=self._get_optuna_pruner()
            )

    def _get_optuna_pruner(self):
        """Returns Optuna's pruning strategy."""
        return optuna.pruners.MedianPruner()

    def save_contour_plot_as_png(self, study):
        """
        Save the contour plot of hyperparameter interactions as a PNG, 
        automatically extracting parameter names from the study object.

        Args:
            study: The Optuna study object.
            output_path: Path to save the PNG file.
        """
        # Extract all parameter names from the study trials
        all_params = set()
        for trial in study.trials:
            all_params.update(trial.params.keys())
        all_params = list(all_params)  # Convert to a sorted list for consistency

        # Generate the contour plot
        fig = optuna.visualization.plot_contour(study, params=all_params) 

        # Adjust figure size and font size
        fig.update_layout(
            width=6000, height=6000,  # Large figure size
            font=dict(size=40)  # Increase font size for better readability
        )

        # Save the plot as a PNG file
        fig.write_image(f'explore_results_{self.model_type}/contour_plot.png', scale=1)  

        # Extract trial data
        trials = [
            {**trial.params, 'objective_value': trial.value}
            for trial in study.trials if trial.state == optuna.trial.TrialState.COMPLETE
        ]

        # Convert to DataFrame
        df = pd.DataFrame(trials)

        # Save to CSV
        df.to_csv(f"explore_results_{self.model_type}/optuna_results.csv", index=False)



from octopi import visualization_tools as viz
from monai.inferers import sliding_window_inference
from octopi import stopping_criteria
from monai.transforms import AsDiscrete
from monai.data import decollate_batch
import torch, os, mlflow, re
import torch_ema as ema
from tqdm import tqdm 
import numpy as np

# Not Ideal, but Necessary if Class is Missing From Dataset
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

class ModelTrainer:

    def __init__(self, 
                 model, 
                 device,
                 loss_function, 
                 metrics_function, 
                 optimizer, 
                 use_ema: bool = True):

        self.model = model
        self.device = device
        self.loss_function = loss_function
        self.metrics_function = metrics_function
        self.optimizer = optimizer

        self.parallel_mlflow = False
        self.client = None
        self.trial_run_id = None  

        # Default F-Beta Value
        self.beta = 2

        # Initialize EMAHandler for the model
        self.ema_experiment = use_ema
        if self.ema_experiment:
            self.ema_handler = ema.ExponentialMovingAverage(self.model.parameters(), decay=0.99)

        # Initialize Figure and Axes for Plotting
        self.fig = None; self.axs = None

    def set_parallel_mlflow(self, 
                            client,
                            trial_run_id):
        
        self.parallel_mlflow = True
        self.client = client
        self.trial_run_id = trial_run_id
    
    def train_update(self):

        step = 0
        epoch_loss = 0
        self.model.train()
        for batch_data in self.train_loader:
            step += 1
            inputs = batch_data["image"].to(self.device)  # Shape: [B, C, H, W, D]
            labels = batch_data["label"].to(self.device)  # Shape: [B, C, H, W, D]
            self.optimizer.zero_grad()
            outputs = self.model(inputs)    # Output shape: [B, num_classes, H, W, D] 
            loss = self.loss_function(outputs, labels)
            loss.backward()
            self.optimizer.step()

            # Update EMA weights
            if self.ema_experiment:
                self.ema_handler.update()

            # Update running epoch loss
            epoch_loss += loss.item()
        
        # Compute and log average epoch loss
        epoch_loss /= step
        return epoch_loss

    def validate_update(self):
        """
        Perform validation and compute metrics, including validation loss.
        """        

        # Set model to evaluation mode
        self.model.eval()
        val_loss = 0
        with torch.no_grad():
            for val_data in self.val_loader:
                val_inputs = val_data["image"].to(self.device)
                val_labels = val_data["label"].to(self.device)
                
                # Apply sliding window inference
                # roi_size=self.input_dim, # try setting a set size of 128, 144 or 160? 
                val_outputs = sliding_window_inference(
                    inputs=val_inputs, 
                    roi_size=(144,144,144),
                    sw_batch_size=4,
                    predictor=self.model, 
                    overlap=0.5,
                    device=self.device
                )

                del val_inputs
                torch.cuda.empty_cache()

                # Compute the loss for this batch
                loss = self.loss_function(val_outputs, val_labels)  # Assuming self.loss_function is defined
                val_loss += loss.item()  # Accumulate the loss                

                # Apply post-processing
                metric_val_outputs = [self.post_pred(i) for i in decollate_batch(val_outputs)]
                metric_val_labels = [self.post_label(i) for i in decollate_batch(val_labels)]                             
                
                # Compute metrics
                self.metrics_function(y_pred=metric_val_outputs, y=metric_val_labels)             

                del val_labels, val_outputs, metric_val_outputs, metric_val_labels
                torch.cuda.empty_cache()

        # # Contains recall, precision, and f1 for each class
        metric_values = self.metrics_function.aggregate(reduction='mean_batch')

        # Compute average validation loss and add to metrics dictionary
        val_loss /= len(self.val_loader)
        metric_values.append(val_loss)

        return metric_values

    def train(
        self,
        data_load_gen,
        model_save_path: str = 'results',
        my_num_samples: int = 15,
        crop_size: int = 96,
        max_epochs: int = 100,
        val_interval: int = 15,
        lr_scheduler_type: str = 'cosine', 
        best_metric: str = 'avg_f1',
        use_mlflow: bool = False,
        verbose: bool = False
    ):

        # best lr scheduler options are cosine or reduce
        self.warmup_epochs = 5
        self.warmup_lr_factor = 0.1
        self.min_lr = 1e-6

        self.max_epochs = max_epochs
        self.crop_size = crop_size
        self.num_samples = my_num_samples
        self.val_interval = val_interval
        self.use_mlflow = use_mlflow

        # Create Save Folder if It Doesn't Exist
        if model_save_path is not None:
            os.makedirs(model_save_path, exist_ok=True)  

        Nclass = data_load_gen.Nclasses
        self.create_results_dictionary(Nclass)  

        # Resolve the best metric
        best_metric = self.resolve_best_metric(best_metric)

        # Stopping Criteria
        self.stopping_criteria = stopping_criteria.EarlyStoppingChecker(monitor_metric=best_metric, val_interval=val_interval)            

        self.post_pred = AsDiscrete(argmax=True, to_onehot=Nclass)
        self.post_label = AsDiscrete(to_onehot=Nclass)                             

        # Produce Dataloaders for the First Training Iteration
        self.train_loader, self.val_loader = data_load_gen.create_train_dataloaders(crop_size=crop_size, num_samples=my_num_samples)
        self.input_dim = data_load_gen.input_dim

        # Save the original learning rate
        original_lr = self.optimizer.param_groups[0]['lr']
        self.load_learning_rate_scheduler(lr_scheduler_type)

        # Initialize tqdm around the epoch loop
        for epoch in tqdm(range(max_epochs), desc="Training Progress", unit="epoch"):

            # Reload dataloaders periodically
            if data_load_gen.reload_frequency > 0 and (epoch + 1) % data_load_gen.reload_frequency == 0:
                self.train_loader, self.val_loader = data_load_gen.create_train_dataloaders(num_samples=my_num_samples)
                # Lower the learning rate for the warm-up period
                for param_group in self.optimizer.param_groups:
                    param_group['lr'] = original_lr * self.warmup_lr_factor

            # Compute and log average epoch loss           
            epoch_loss = self.train_update()

            # Check for NaN in the loss
            if self.stopping_criteria.should_stop_training(epoch_loss):
                tqdm.write(f"Training stopped early due to {self.stopping_criteria.get_stopped_reason()}")
                break

            current_lr = self.optimizer.param_groups[0]['lr']
            self.my_log_metrics( metrics_dict={"loss": epoch_loss}, curr_step=epoch + 1 )
            self.my_log_metrics( metrics_dict={"learning_rate": current_lr}, curr_step=epoch + 1 )                        

            # Validation and metric logging
            if (epoch + 1) % val_interval == 0 or (epoch + 1) == max_epochs:
                if verbose:
                    tqdm.write(f"Epoch {epoch + 1}/{max_epochs}, avg_train_loss: {epoch_loss:.4f}")

                # Validate the Model with or without EMA
                if self.ema_experiment:
                    with self.ema_handler.average_parameters():
                        metric_values = self.validate_update()
                else:
                    metric_values = self.validate_update()

                # Log all metrics
                self.my_log_metrics( metrics_dict=metric_values, curr_step=epoch + 1 )

                # Update tqdm description        
                if verbose:
                    (avg_f1, avg_recall, avg_precision) = (self.results['avg_f1'][-1][1], 
                                                           self.results['avg_recall'][-1][1], 
                                                           self.results['avg_precision'][-1][1])
                    tqdm.write(f"Epoch {epoch + 1}/{max_epochs}, avg_f1_score: {avg_f1:.4f}, avg_recall: {avg_recall:.4f}, avg_precision: {avg_precision:.4f}")

                # Reset metrics function
                self.metrics_function.reset()

                # Save the best model
                if self.results[best_metric][-1][1] > self.results["best_metric"]:
                    self.results["best_metric"] = self.results[best_metric][-1][1]
                    self.results["best_metric_epoch"] = epoch + 1

                    # Read Model Weights and Save
                    if self.ema_experiment:
                        with self.ema_handler.average_parameters():
                            self.save_model(model_save_path)
                    else:
                        self.save_model(model_save_path)

                # Save plot if Local Training Call
                if not self.use_mlflow:
                    self.fig, self.axs = viz.plot_training_results(
                        self.results, 
                        save_plot=os.path.join(model_save_path, "net_train_history.png"), 
                        fig=self.fig, 
                        axs=self.axs)

                # After Validation Metrics are Logged, Check for Early Stopping
                if self.stopping_criteria.should_stop_training(epoch_loss, results=self.results, check_metrics=True):
                    tqdm.write(f"Training stopped early due to {self.stopping_criteria.get_stopped_reason()}")
                    break

            # Run the learning rate scheduler
            early_stop = self.run_scheduler(data_load_gen, original_lr, epoch, val_interval, lr_scheduler_type)
            if early_stop:
                break

        return self.results

    def load_learning_rate_scheduler(self, type: str = 'cosine'):
        """
        Initialize and return the learning rate scheduler based on the given type.
        """
        # Configure learning rate scheduler based on the type
        if type == "cosine":
            self.lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                self.optimizer, T_max=self.max_epochs, eta_min=self.min_lr )
        elif type == "onecyle":
            max_lr = 1e-3
            steps_per_epoch = len(self.train_loader)
            self.lr_scheduler = torch.optim.lr_scheduler.OneCycleLR(
                self.optimizer, max_lr=max_lr, epochs=self.max_epochs, steps_per_epoch=steps_per_epoch )
        elif type == "reduce":
            mode = "min"
            patience = 3
            factor = 0.5
            self.lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, mode=mode, patience=patience, factor=factor )
        elif type == 'exponential':
            gamma = 0.9
            self.lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(
                self.optimizer, gamma=gamma)
        else:
            raise ValueError(f"Unsupported scheduler type: {type}")

    def run_scheduler(
        self, 
        data_load_gen, 
        original_lr: float,
        epoch: int,
        val_interval: int,
        type: str
        ):
        """
        Manage the learning rate scheduler, including warm-up and normal scheduling.
        """
        # Apply warm-up logic
        if (epoch + 1) <= self.warmup_epochs:
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = original_lr * self.warmup_lr_factor
            return False  # Continue training

        # Step the scheduler
        if isinstance(self.lr_scheduler, torch.optim.lr_scheduler.CosineAnnealingLR):
            self.lr_scheduler.step()
        elif isinstance(self.lr_scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau) and (epoch + 1) % val_interval == 0:
            metric_value = self.results['val_loss'][-1][1]
            self.lr_scheduler.step(metric_value)
        else:
            self.lr_scheduler.step()  # Step for other schedulers

        # Check learning rate for early stopping
        current_lr = self.optimizer.param_groups[0]['lr']
        if current_lr < self.min_lr and type != 'onecycle':
            print(f"Early stopping triggered at epoch {epoch + 1} as learning rate fell below {self.min_lr}.")
            return True  # Indicate early stopping

        return False  # Continue training
        
    def save_model(self, model_save_path: str):

        # Store Model Weights as Member Variable
        self.model_weights = self.model.state_dict()

        # Save Model Weights to *.pth file
        if model_save_path is not None:
            torch.save(self.model_weights, os.path.join(model_save_path, "best_model.pth"))

    def create_results_dictionary(self, Nclass: int):

        self.results = {
            'loss': [],
            'val_loss': [],
            'avg_f1': [],
            'avg_recall': [],
            'avg_precision': [],
            'avg_fbeta': [],
            'best_metric': -1,  # Initialize as None or a default value
            'best_metric_epoch': -1
        }

        for i in range(Nclass-1):
            self.results[f'fbeta_class{i+1}'] = []
            self.results[f'f1_class{i+1}'] = []
            self.results[f'recall_class{i+1}'] = []
            self.results[f'precision_class{i+1}'] = []

        self.metric_names = self.results.keys()

    def my_log_metrics(
        self,
        metrics_dict: dict,
        curr_step: int,
        ):

        # If metrics_dict contains multiple elements (e.g., recall, precision, f1), process them
        if len(metrics_dict) > 1:

            # Extract individual metrics 
            # (assume metrics_dict contains recall, precision, f1, val_loss in sequence)
            recall, precision, f1s, val_loss = metrics_dict[0], metrics_dict[1], metrics_dict[2], metrics_dict[3]

            # Log per-class metrics
            metrics_to_log = {}
            for i, (rec, prec, f1) in enumerate(zip(recall, precision, f1s)):
                metrics_to_log[f"recall_class{i+1}"] = rec.item()
                metrics_to_log[f"precision_class{i+1}"] = prec.item()
                metrics_to_log[f"f1_class{i+1}"] = f1.item()
                metrics_to_log[f"fbeta_class{i+1}"] = self.fbeta(prec, rec).item()

            # Prepare average metrics
            metrics_to_log["avg_recall"] = recall.mean().cpu().item()
            metrics_to_log["avg_precision"] = precision.mean().cpu().item()
            metrics_to_log["avg_f1"] = f1s.mean().cpu().item()
            metrics_to_log["avg_fbeta"] = self.fbeta(precision, recall).mean().cpu().item()
            metrics_to_log['val_loss'] = val_loss

            # Update metrics_dict for further logging
            metrics_dict = metrics_to_log

        # Log all metrics (per-class and average metrics)
        for metric_name, value in metrics_dict.items():
            if metric_name not in self.results:
                self.results[metric_name] = []
            self.results[metric_name].append((curr_step, value))

        # Log to MLflow or client
        if self.client is not None and self.trial_run_id is not None:
            for metric_name, value in metrics_dict.items():
                self.client.log_metric(
                    run_id=self.trial_run_id,
                    key=metric_name,
                    value=value,
                    step=curr_step,
                )
        elif self.use_mlflow:
            for metric_name, value in metrics_dict.items():
                mlflow.log_metric(metric_name, value, step=curr_step)

    def fbeta(self, precision, recall):

        # Handle division by zero
        numerator = (1 + self.beta**2) * (precision * recall)
        denominator = (self.beta**2 * precision) + recall
        
        # Use torch.where to handle zero cases
        result = torch.where(
            denominator > 0,
            numerator / denominator,
            torch.zeros_like(precision)
        )
        return result

    def my_log_params(
        self,
        params_dict: dict, 
        ):

        if self.client is not None and self.trial_run_id is not None:
            for key, value in params_dict.items():
                self.client.log_param(run_id=self.trial_run_id, key=key, value=value)
        else:
            mlflow.log_params(params_dict)  

    # Example input: best_metric = 'fBeta2_class3' or 'fBeta1' or 'f1_class2'
    def resolve_best_metric(self, best_metric):
        fbeta_pattern = r"^fBeta(\d+)(?:_class(\d+))?$"  # Matches fBetaX or fBetaX_classY
        match = re.match(fbeta_pattern, best_metric)

        if match:
            self.beta = int(match.group(1))  # Extract beta value
            class_part = match.group(2)
            if class_part:
                best_metric = f'fbeta_class{class_part}'  # fBeta2_class3 → fbeta_class3
            else:
                best_metric = 'avg_fbeta'  # fBeta2 → avg_fbeta

        elif best_metric in self.metric_names:
            pass  # It's already a valid metric in the results dict

        else:
            print(f"'{best_metric}' is not a valid metric. Defaulting to 'avg_f1'.\n")
            best_metric = 'avg_f1'

        return best_metric
    



from monai.transforms import (
    Compose, 
    RandFlipd, 
    Orientationd, 
    RandRotate90d, 
    NormalizeIntensityd,
    EnsureChannelFirstd, 
    RandCropByLabelClassesd,
    RandScaleIntensityd,
    RandShiftIntensityd,
    RandAdjustContrastd,
    RandGaussianNoised,
    ScaleIntensityRanged,  
    RandomOrder,
)

def get_transforms():
    """
    Returns non-random transforms.
    """
    return Compose([
        EnsureChannelFirstd(keys=["image", "label"], channel_dim="no_channel"),
        NormalizeIntensityd(keys="image"),
        Orientationd(keys=["image", "label"], axcodes="RAS")
    ])

def get_random_transforms( input_dim, num_samples, Nclasses):
    """
    Input:
        input_dim: tuple of (nx, ny, nz)
        num_samples: int
        Nclasses: int

    Returns random transforms.
    
    For data with a missing wedge along the first axis (causing smearing in that direction),
    we avoid rotations that would move this artifact to other axes. We only rotate around
    the first axis (spatial_axes=[1, 2]) and avoid flipping along the first axis.
    """
    return Compose([
        RandCropByLabelClassesd(
            keys=["image", "label"],
            label_key="label",
            spatial_size=[input_dim[0], input_dim[1], input_dim[2]],     
            num_classes=Nclasses,
            num_samples=num_samples
        ),
        # Only rotate around the first axis (keeping the missing wedge orientation consistent)
        RandRotate90d(keys=["image", "label"], prob=0.5, spatial_axes=[1, 2], max_k=3),
        # Avoid flipping along the first axis (where the missing wedge is)
        RandFlipd(keys=["image", "label"], prob=0.5, spatial_axis=0),  # Removed
        RandFlipd(keys=["image", "label"], prob=0.5, spatial_axis=1),
        RandFlipd(keys=["image", "label"], prob=0.5, spatial_axis=2),        
        RandomOrder([
            # Intensity augmentations are still appropriate
            RandScaleIntensityd(keys="image", prob=0.5, factors=(0.85, 1.15)),
            RandShiftIntensityd(keys="image", prob=0.5, offsets=(-0.15, 0.15)),
            RandAdjustContrastd(keys="image", prob=0.5, gamma=(0.85, 1.15)),
            RandGaussianNoised(keys="image", prob=0.5, mean=0.0, std=0.5),  # Reduced noise std
        ]),
    ]) 

    # Augmentations to Explore in the Future: 
    # Intensity-based augmentations
    # RandHistogramShiftd(keys="image", prob=0.5, num_control_points=(3, 5))
    # RandGaussianSmoothd(keys="image", prob=0.5, sigma_x=(0.5, 1.5), sigma_y=(0.5, 1.5), sigma_z=(0.5, 1.5)),

    # Geometric Transforms
    # RandAffined(
    #     keys=["image", "label"],
    #     rotate_range=(0.1, 0.1, 0.1),  # Rotation angles (radians) for x, y, z axes
    #     scale_range=(0.1, 0.1, 0.1),   # Scale range for isotropic/anisotropic scaling
    #     prob=0.5,                      # Probability of applying the transform
    #     padding_mode="border"          # Handle out-of-bounds values
    # )    

def get_predict_transforms():
    """
    Returns predict transforms.
    """
    return Compose([
        EnsureChannelFirstd(keys=["image"], channel_dim="no_channel"),
        NormalizeIntensityd(keys="image")
    ])



from monai.losses import FocalLoss, TverskyLoss
import torch

class WeightedFocalTverskyLoss(torch.nn.Module):
    def __init__(
        self, gamma=1.0, alpha=0.7, beta=0.3, 
        weight_tversky=0.5, weight_focal=0.5, 
        smooth=1e-5, **kwargs ):
        """
        Weighted combination of Focal and Tversky loss.

        Args:
            gamma (float): Focus parameter for Focal Loss.
            alpha (float): Weight for false positives in Tversky Loss.
            beta (float): Weight for false negatives in Tversky Loss.
            weight_tversky (float): Weight of Tversky loss in the combination.
            weight_focal (float): Weight of Focal loss in the combination.
            smooth (float): Smoothing factor to avoid division by zero.
        """
        super().__init__()
        self.tversky_loss = TverskyLoss(
            alpha=alpha, beta=beta, include_background=True, 
            to_onehot_y=True, softmax=True,
            smooth_nr=smooth, smooth_dr=smooth, **kwargs
        )
        self.focal_loss = FocalLoss(
            include_background=True, to_onehot_y=True, 
            use_softmax=True, gamma=gamma
        )
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.weight_tversky = weight_tversky
        self.weight_focal = weight_focal

    def forward(self, y_pred, y_true):
        """
        Compute the combined loss.

        Args:
            y_pred (Tensor): Predicted probabilities (B, C, ...).
            y_true (Tensor): Ground truth labels (B, C, ...), one-hot encoded.

        Returns:
            Tensor: Weighted combination of Tversky and Focal losses.
        """
        tversky = self.tversky_loss(y_pred, y_true)
        focal = self.focal_loss(y_pred, y_true)
        return self.weight_tversky * tversky + self.weight_focal * focal
    
class FocalTverskyLoss(TverskyLoss):
    def __init__(
        self, 
        alpha=0.7, beta=0.3, gamma=1.0, smooth=1e-5, **kwargs):
        """
        Focal Tversky Loss with an additional power term for harder samples.

        From https://arxiv.org/abs/1810.07842

        Args:
            alpha (float): Weight for false positives.
            beta (float): Weight for false negatives.
            gamma (float): Focus parameter (like Focal Loss).
            smooth (float): Smoothing factor to avoid division by zero.
        """
        super().__init__(
            alpha=alpha, beta=beta, 
            include_background=True, 
            to_onehot_y=True, softmax=True, 
            smooth_nr=smooth, smooth_dr=smooth, **kwargs)
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta

    def forward(self, y_pred, y_true):
        """
        Args:
            y_pred (Tensor): Predicted probabilities (B, C, ...).
            y_true (Tensor): Ground truth labels (B, C, ...), one-hot encoded.

        Returns:
            Tensor: Loss value.
        """
        tversky_loss = super().forward(y_pred, y_true)
        modified_loss = torch.pow(tversky_loss, 1 / self.gamma)
        return modified_loss



from monai.losses import FocalLoss, TverskyLoss
from octopi import losses
from octopi.models import (
    Unet, AttentionUnet, MedNeXt, SegResNet
)

def get_model(architecture):

    # Initialize model based on architecture
    if architecture == "Unet":
        model = Unet.myUNet()
    elif architecture == "AttentionUnet":
        model = AttentionUnet.myAttentionUnet()
    elif architecture == "MedNeXt":
        model = MedNeXt.myMedNeXt()
    elif architecture == "SegResNet":
        model = SegResNet.mySegResNet()
    else:
        raise ValueError(f"Model type {architecture} not supported!\nPlease use one of the following: Unet, AttentionUnet, MedNeXt, SegResNet")

    return model

def get_loss_function(trial, loss_name = None):

    # Loss function selection
    if loss_name is None:
        loss_name = trial.suggest_categorical(
            "loss_function", 
            ["FocalLoss", "WeightedFocalTverskyLoss", 'FocalTverskyLoss'])

    if loss_name == "FocalLoss":
        gamma = round(trial.suggest_float("gamma", 0.1, 2), 3)
        loss_function = FocalLoss(include_background=True, to_onehot_y=True, use_softmax=True, gamma=gamma)

    elif loss_name == "TverskyLoss":
        alpha = round(trial.suggest_float("alpha", 0.1, 0.5), 3)
        beta = 1.0 - alpha
        loss_function = TverskyLoss(include_background=True, to_onehot_y=True, softmax=True, alpha=alpha, beta=beta)

    elif loss_name == 'WeightedFocalTverskyLoss':
        gamma = round(trial.suggest_float("gamma", 0.1, 2), 3)
        alpha = round(trial.suggest_float("alpha", 0.1, 0.5), 3)
        beta = 1.0 - alpha
        weight_tversky = round(trial.suggest_float("weight_tversky", 0.1, 0.9), 3)
        weight_focal = 1.0 - weight_tversky
        loss_function = losses.WeightedFocalTverskyLoss(
            gamma=gamma, alpha=alpha, beta=beta,
            weight_tversky=weight_tversky, weight_focal=weight_focal
        )

    elif loss_name == 'FocalTverskyLoss':
        gamma = round(trial.suggest_float("gamma", 0.1, 2), 3)
        alpha = round(trial.suggest_float("alpha", 0.1, 0.5), 3)
        beta = 1.0 - alpha
        loss_function = losses.FocalTverskyLoss(gamma=gamma, alpha=alpha, beta=beta)

    return loss_function


#### TODO : Models to try Adding? 
# 1. Swin UNETR 
# 2. Swin-Conv-UNet


from monai.data import DataLoader, CacheDataset, Dataset
from monai.transforms import (
    Compose, 
    NormalizeIntensityd,
    EnsureChannelFirstd,  
)
from sklearn.model_selection import train_test_split
import copick, torch, os, json, random
from collections import defaultdict
from octopi import utils
from typing import List
from tqdm import tqdm
import numpy as np

##############################################################################################################################    

def load_training_data(root, 
                       runIDs: List[str],
                       voxel_spacing: float, 
                       tomo_algorithm: str, 
                       segmenation_name: str,
                       segmentation_session_id: str = None,
                       segmentation_user_id: str = None,
                       progress_update: bool = True):
    
    data_dicts = []
    # Use tqdm for progress tracking only if progress_update is True
    iterable = tqdm(runIDs, desc="Loading Training Data") if progress_update else runIDs
    for runID in iterable:
        run = root.get_run(str(runID))
        tomogram = get_tomogram_array(run, voxel_spacing, tomo_algorithm)
        segmentation = get_segmentation_array(run, 
                                              voxel_spacing,
                                              segmenation_name,
                                              segmentation_session_id, 
                                              segmentation_user_id)
        data_dicts.append({"image": tomogram, "label": segmentation})

    return data_dicts 

##############################################################################################################################    

def load_predict_data(root, 
                      runIDs: List[str],
                      voxel_spacing: float, 
                      tomo_algorithm: str):
    
    data_dicts = []
    for runID in tqdm(runIDs):
        run = root.get_run(str(runID))
        tomogram = get_tomogram_array(run, voxel_spacing, tomo_algorithm)
        data_dicts.append({"image": tomogram})

    return data_dicts 

##############################################################################################################################    

def create_predict_dataloader(
    root,
    voxel_spacing: float, 
    tomo_algorithm: str,       
    runIDs: str = None,       
    ): 

    # define pre transforms
    pre_transforms = Compose(
        [   EnsureChannelFirstd(keys=["image"], channel_dim="no_channel"),
            NormalizeIntensityd(keys=["image"]),
    ])

    # Split trainRunIDs, validateRunIDs, testRunIDs
    if runIDs is None:
        runIDs = [run.name for run in root.runs]
    test_files = load_predict_data(root, runIDs, voxel_spacing, tomo_algorithm) 

    bs = min( len(test_files), 4)
    test_ds = CacheDataset(data=test_files, transform=pre_transforms)
    test_loader = DataLoader(test_ds, 
                            batch_size=bs, 
                            shuffle=False, 
                            num_workers=4, 
                            pin_memory=torch.cuda.is_available())
    return test_loader, test_ds

##############################################################################################################################

def get_tomogram_array(run, 
                       voxel_size: float = 10, 
                       tomo_type: str = 'wbp',
                       raise_error: bool = True):
    
    voxel_spacing_obj = run.get_voxel_spacing(voxel_size)

    if voxel_spacing_obj is None:
        # Query Avaiable Voxel Spacings
        availableVoxelSpacings = [tomo.voxel_size for tomo in run.voxel_spacings]

        # Report to the user which voxel spacings they can use 
        message = (f"\n[Warning] No tomogram found for {run.name} with voxel size {voxel_size} and tomogram type {tomo_type}"
                   f"\nAvailable spacings are: {', '.join(map(str, availableVoxelSpacings))}\n" ) 
        if raise_error:
            raise ValueError(message)
        else:
            print(message)
            return None
    
    tomogram = voxel_spacing_obj.get_tomogram(tomo_type)
    if tomogram is None:
        # Get available algorithms
        availableAlgorithms = [tomo.tomo_type for tomo in run.get_voxel_spacing(voxel_size).tomograms]
        
        # Report to the user which algorithms are available
        message = (f"\n[Warning] No tomogram found for {run.name} with voxel size {voxel_size} and tomogram type {tomo_type}"
                   f"\nAvailable algorithms are: {', '.join(availableAlgorithms)}\n")
        if raise_error:
            raise ValueError(message)
        else:
            print(message)
            return None

    return tomogram.numpy().astype(np.float32)

##############################################################################################################################

def get_segmentation_array(run, 
                           voxel_spacing: float,
                           segmentation_name: str, 
                           session_id=None,
                           user_id=None,
                           raise_error: bool = True):

    seg = run.get_segmentations(name=segmentation_name, 
                                session_id = session_id,
                                user_id = user_id,
                                voxel_size=float(voxel_spacing))
    
    # No Segmentations Are Available, Result in Error
    if len(seg) == 0:
        # Get all available segmentations with their metadata
        available_segs = run.get_segmentations(voxel_size=float(voxel_spacing))
        seg_info = [(s.name, s.user_id, s.session_id) for s in available_segs]
        
        # Format the information for display
        seg_details = [f"(name: {name}, user_id: {uid}, session_id: {sid})" 
                      for name, uid, sid in seg_info]
        
        message = ( f'\nNo segmentation found matching:\n'
                    f'  name: {segmentation_name}, user_id: {user_id}, session_id: {session_id}\n'
                    f'Available segmentations in {run.name} are:\n  ' + 
                    '\n  '.join(seg_details) )
        if raise_error:
            raise ValueError(message)
        else:
            print(message)
            return None

    # No Segmentations Are Available, Result in Error
    if len(seg) > 1:
        print(f'[Warning] More Than 1 Segmentation is Available for the Query Information. '
              f'Available Segmentations are: {seg} '
              f'Defaulting to Loading: {seg[0]}\n')
    seg = seg[0]

    return seg.numpy().astype(np.int8)

##############################################################################################################################

def get_copick_coordinates(run,                    # CoPick run object containing the segmentation data
                           name: str,              # Name of the object or protein for which coordinates are being extracted
                           user_id: str,           # Identifier of the user that generated the picks
                           session_id: str = None, # Identifier of the session that generated the picks
                           voxel_size: float = 10,  # Voxel size of the tomogram, used for scaling the coordinates
                           raise_error: bool = True):
                           
    # Retrieve the pick points associated with the specified object and user ID
    picks = run.get_picks(object_name=name, user_id=user_id, session_id=session_id)

    if len(picks) == 0:
        # Get all available segmentations with their metadata

        available_picks = run.get_picks()
        picks_info = [(s.pickable_object_name, s.user_id, s.session_id) for s in available_picks]
        
        # Format the information for display
        picks_details = [f"(name: {name}, user_id: {uid}, session_id: {sid})" 
                      for name, uid, sid in picks_info]
        
        message = ( f'\nNo picks found matching:\n'
                    f'  name: {name}, user_id: {user_id}, session_id: {session_id}\n'
                    f'Available picks are:\n  '  
                    + '\n  '.join(picks_details) )
        if raise_error:
            raise ValueError(message)
        else:
            print(message)
            return None
    elif len(picks) > 1:
        # Format pick information for display
        picks_info = [(p.pickable_object_name, p.user_id, p.session_id) for p in picks]
        picks_details = [f"(name: {name}, user_id: {uid}, session_id: {sid})" 
                        for name, uid, sid in picks_info]

        print(f'[Warning] More than 1 pick is available for the query information.'
              f'\nAvailable picks are:\n  ' + 
              '\n  '.join(picks_details) +
              f'\nDefaulting to loading:\n {picks[0]}\n')
    points = picks[0].points

    # Initialize an array to store the coordinates
    nPoints = len(picks[0].points)                      # Number of points retrieved
    coordinates = np.zeros([len(picks[0].points), 3])   # Create an empty array to hold the (z, y, x) coordinates

    # Iterate over all points and convert their locations to coordinates in voxel space
    for ii in range(nPoints):
        coordinates[ii,] = [points[ii].location.z / voxel_size,   # Scale z-coordinate by voxel size
                            points[ii].location.y / voxel_size,   # Scale y-coordinate by voxel size
                            points[ii].location.x / voxel_size]   # Scale x-coordinate by voxel size

    # Return the array of coordinates
    return coordinates
    

##############################################################################################################################

def adjust_to_multiple(value, multiple = 16):
    return int((value // multiple) * multiple)

def get_input_dimensions(dataset, crop_size: int):
    nx = dataset[0]['image'].shape[1]
    if crop_size > nx:
        first_dim = adjust_to_multiple(nx/2)
        return first_dim, crop_size, crop_size
    else:
        return crop_size, crop_size, crop_size

def get_num_classes(copick_config_path: str):

    root = copick.from_file(copick_config_path)
    return len(root.pickable_objects) + 1

def split_multiclass_dataset(runIDs, 
                             train_ratio: float = 0.7, 
                             val_ratio: float = 0.15, 
                             test_ratio: float = 0.15, 
                             return_test_dataset: bool = True,
                             random_state: int = 42):
    """
    Splits a given dataset into three subsets: training, validation, and testing. If the dataset
    has categories (as tuples), splits are balanced across all categories. If the dataset is a 1D
    list, it is split without categorization.

    Parameters:
    - runIDs: A list of items to split. It can be a 1D list or a list of tuples (category, value).
    - train_ratio: Proportion of the dataset for training.
    - val_ratio: Proportion of the dataset for validation.
    - test_ratio: Proportion of the dataset for testing.
    - return_test_dataset: Whether to return the test dataset.
    - random_state: Random state for reproducibility.

    Returns:
    - trainRunIDs: Training subset.
    - valRunIDs: Validation subset.
    - testRunIDs: Testing subset (if return_test_dataset is True, otherwise None).
    """

    # Ensure the ratios add up to 1
    assert train_ratio + val_ratio + test_ratio == 1.0, "Ratios must add up to 1.0"

    # Check if the dataset has categories
    if isinstance(runIDs[0], tuple) and len(runIDs[0]) == 2:
        # Group by category
        grouped = defaultdict(list)
        for item in runIDs:
            grouped[item[0]].append(item)

        # Split each category
        trainRunIDs, valRunIDs, testRunIDs = [], [], []
        for category, items in grouped.items():
            # Shuffle for randomness
            random.shuffle(items)
            # Split into train and remaining
            train_items, remaining = train_test_split(items, test_size=(1 - train_ratio), random_state=random_state)
            trainRunIDs.extend(train_items)

            if return_test_dataset:
                # Split remaining into validation and test
                val_items, test_items = train_test_split(
                    remaining,
                    test_size=(test_ratio / (val_ratio + test_ratio)),
                    random_state=random_state,
                )
                valRunIDs.extend(val_items)
                testRunIDs.extend(test_items)
            else:
                valRunIDs.extend(remaining)
                testRunIDs = []
    else:
        # If no categories, split as a 1D list
        trainRunIDs, remaining = train_test_split(runIDs, test_size=(1 - train_ratio), random_state=random_state)
        if return_test_dataset:
            valRunIDs, testRunIDs = train_test_split(
                remaining,
                test_size=(test_ratio / (val_ratio + test_ratio)),
                random_state=random_state,
            )
        else:
            valRunIDs = remaining
            testRunIDs = []

    return trainRunIDs, valRunIDs, testRunIDs    

##############################################################################################################################

def load_copick_config(path: str):

    if os.path.isfile(path):
        root = copick.from_file(path)
    else:
        raise FileNotFoundError(f"Copick Config Path does not exist: {path}")
    
    return root

##############################################################################################################################

# Helper function to flatten and serialize nested parameters
def flatten_params(params, parent_key=''):
    flattened = {}
    for key, value in params.items():
        new_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, dict):
            flattened.update(flatten_params(value, new_key))
        elif isinstance(value, list):
            flattened[new_key] = ', '.join(map(str, value))  # Convert list to a comma-separated string
        else:
            flattened[new_key] = value
    return flattened

# Manually join specific lists into strings for inline display
def prepare_for_inline_json(data):
    for key in ["trainRunIDs", "valRunIDs", "testRunIDs"]:
        if key in data['dataloader']:
            data['dataloader'][key] = f"[{', '.join(map(repr, data['dataloader'][key]))}]"

    for key in ['channels', 'strides']:
        if key in data['model']:
                data['model'][key] = f"[{', '.join(map(repr, data['model'][key]))}]"
    return data

def get_optimizer_parameters(trainer):

    optimizer_parameters = {
        'my_num_samples': trainer.num_samples,  
        'val_interval': trainer.val_interval,
        'lr': trainer.optimizer.param_groups[0]['lr'],
        'optimizer': trainer.optimizer.__class__.__name__,
        'metrics_function': trainer.metrics_function.__class__.__name__,
        'loss_function': trainer.loss_function.__class__.__name__,
    }

    # Log Tversky Loss Parameters
    if trainer.loss_function.__class__.__name__ == 'TverskyLoss':
        optimizer_parameters['alpha'] = trainer.loss_function.alpha
    elif trainer.loss_function.__class__.__name__ == 'FocalLoss':
        optimizer_parameters['gamma'] = trainer.loss_function.gamma
    elif trainer.loss_function.__class__.__name__ == 'WeightedFocalTverskyLoss':
        optimizer_parameters['alpha'] = trainer.loss_function.alpha
        optimizer_parameters['gamma'] = trainer.loss_function.gamma
        optimizer_parameters['weight_tversky'] = trainer.loss_function.weight_tversky
    elif trainer.loss_function.__class__.__name__ == 'FocalTverskyLoss':
        optimizer_parameters['alpha'] = trainer.loss_function.alpha
        optimizer_parameters['gamma'] = trainer.loss_function.gamma

    return optimizer_parameters

def save_parameters_to_yaml(model, trainer, dataloader, filename: str):

    parameters = {
        'model': model.get_model_parameters(),
        'optimizer': get_optimizer_parameters(trainer),
        'dataloader': dataloader.get_dataloader_parameters()
    }

    utils.save_parameters_yaml(parameters, filename)
    print(f"Training Parameters saved to {filename}")

def prepare_inline_results_json(results):
    # Traverse the dictionary and format lists of lists as inline JSON
    for key, value in results.items():
        # Check if the value is a list of lists (like [[epoch, value], ...])
        if isinstance(value, list) and all(isinstance(item, list) and len(item) == 2 for item in value):
            # Format the list of lists as a single-line JSON string
            results[key] = json.dumps(value)
    return results    

# Check to See if I'm Happy with This... Maybe Save as H5 File? 
def save_results_to_json(results, filename: str):

    results = prepare_inline_results_json(results)
    with open(os.path.join(filename), "w") as json_file:
        json.dump( results, json_file, indent=4 )
    print(f"Training Results saved to {filename}")

##############################################################################################################################

# def save_parameters_to_json(model, trainer, dataloader, filename: str):

#     parameters = {
#         'model': model.get_model_parameters(),
#         'optimizer': get_optimizer_parameters(trainer),
#         'dataloader': dataloader.get_dataloader_parameters()
#     }
#     parameters = prepare_for_inline_json(parameters)

#     with open(os.path.join(filename), "w") as json_file:
#         json.dump( parameters, json_file, indent=4 )
#     print(f"Training Parameters saved to {filename}")

# def split_datasets(runIDs, 
#                    train_ratio: float = 0.7, 
#                    val_ratio: float = 0.15, 
#                    test_ratio: float = 0.15, 
#                    return_test_dataset: bool = True,
#                    random_state: int = 42):
#     """
#     Splits a given dataset into three subsets: training, validation, and testing. The proportions
#     of each subset are determined by the provided ratios, ensuring that they add up to 1. The
#     function uses a fixed random state for reproducibility.

#     Parameters:
#     - runIDs: The complete dataset that needs to be split.
#     - train_ratio: The proportion of the dataset to be used for training.
#     - val_ratio: The proportion of the dataset to be used for validation.
#     - test_ratio: The proportion of the dataset to be used for testing.

#     Returns:
#     - trainRunIDs: The subset of the dataset used for training.
#     - valRunIDs: The subset of the dataset used for validation.
#     - testRunIDs: The subset of the dataset used for testing.
#     """

#     # Ensure the ratios add up to 1
#     assert train_ratio + val_ratio + test_ratio == 1.0, "Ratios must add up to 1.0"

#     # First, split into train and remaining (30%)
#     trainRunIDs, valRunIDs = train_test_split(runIDs, test_size=(1 - train_ratio), random_state=random_state)

#     # (Optional) split the remaining into validation and test
#     if return_test_dataset: 
#         valRunIDs, testRunIDs = train_test_split(
#             valRunIDs,
#             test_size=(test_ratio / (val_ratio + test_ratio)),
#             random_state=random_state,
#         )
#     else:
#         testRunIDs = None

#     return trainRunIDs, valRunIDs, testRunIDs



from octopi.entry_points import common
from octopi.extract import localize
from octopi import utils
import copick, argparse, pprint
from typing import List, Tuple
import multiprocess as mp
from tqdm import tqdm
import os

def pick_particles(
    copick_config_path: str,
    method: str,
    seg_info: Tuple[str, str, str],
    voxel_size: float,
    pick_session_id: str,
    pick_user_id: str,
    radius_min_scale: float,
    radius_max_scale: float,
    filter_size: float,
    pick_objects: List[str],
    runIDs: List[str],
    n_procs: int,
    ):

    # Load the Copick Project
    root = copick.from_file(copick_config_path)    

    # Get objects that can be Picked
    objects = [(obj.name, obj.label, obj.radius) for obj in root.pickable_objects if obj.is_particle]

     # Verify each object has the required attributes
    for obj in objects:
        if len(obj) < 3 or not isinstance(obj[2], (float, int)):
            raise ValueError(f"Invalid object format: {obj}. Expected a tuple with (name, label, radius).")

    # Filter elements
    if pick_objects is not None:
        objects = [obj for obj in objects if obj[0] in pick_objects]

    print(f'Running Localization on the Following Objects: ')
    print(', '.join([f'{obj[0]} (Label: {obj[1]})' for obj in objects]) + '\n')

    # Either Specify Input RunIDs or Run on All RunIDs
    if runIDs:
        print('Running Localization on the Following RunIDs: ' + ', '.join(runIDs) + '\n')
        run_ids = runIDs
    else:
        run_ids = [run.name for run in root.runs if run.get_voxel_spacing(voxel_size) is not None]
        skipped_run_ids = [run.name for run in root.runs if run.get_voxel_spacing(voxel_size) is None]

        if skipped_run_ids:
            print(f"Warning: skipping runs with no voxel spacing {voxel_size}: {skipped_run_ids}")

    # Nprocesses shouldnt exceed computation resource or number of available runs
    n_run_ids = len(run_ids)
    n_procs = min(mp.mp.cpu_count(), n_procs, n_run_ids)

    # Run Localization - Main Parallelization Loop
    print(f"Using {n_procs} processes to parallelize across {n_run_ids} run IDs.")
    with mp.Pool(processes=n_procs) as pool:
        with tqdm(total=n_run_ids, desc="Localization", unit="run") as pbar:
            worker_func = lambda run_id: localize.processs_localization(
                root.get_run(run_id),  
                objects, 
                seg_info,
                method, 
                voxel_size,
                filter_size,
                radius_min_scale, 
                radius_max_scale,
                pick_session_id,
                pick_user_id
            )

            for _ in pool.imap_unordered(worker_func, run_ids, chunksize=1):
                pbar.update(1)

    print('Localization Complete!')

def localize_parser(parser_description, add_slurm: bool = False):
    parser = argparse.ArgumentParser(
        description=parser_description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    input_group = parser.add_argument_group("Input Arguments")
    input_group.add_argument("--config", type=str, required=True, help="Path to the CoPick configuration file.")
    input_group.add_argument("--method", type=str, choices=['watershed', 'com'], default='watershed', required=False, help="Localization method to use.")
    input_group.add_argument('--seg-info', type=utils.parse_target, required=True, help='Query for the organelles segmentations (e.g., "name" or "name,user_id,session_id".).')
    input_group.add_argument("--voxel-size", type=float, default=10, required=False, help="Voxel size for localization.")
    input_group.add_argument("--runIDs", type=utils.parse_list, default = None, required=False, help="List of runIDs to run inference on, e.g., run1,run2,run3 or [run1,run2,run3].")

    localize_group = parser.add_argument_group("Localize Arguments")
    localize_group.add_argument("--radius-min-scale", type=float, default=0.5, required=False, help="Minimum radius scale for particles.")
    localize_group.add_argument("--radius-max-scale", type=float, default=1.0, required=False, help="Maximum radius scale for particles.")
    localize_group.add_argument("--filter-size", type=int, default=10, required=False, help="Filter size for localization.")
    localize_group.add_argument("--pick-objects", type=utils.parse_list, default=None, required=False, help="Specific Objects to Find Picks for.")
    localize_group.add_argument("--n-procs", type=int, default=8, required=False, help="Number of CPU processes to parallelize runs across. Defaults to the max number of cores available or available runs.")

    output_group = parser.add_argument_group("Output Arguments")
    output_group.add_argument("--pick-session-id", type=str, default='1', required=False, help="Session ID for the particle picks.")
    output_group.add_argument("--pick-user-id", type=str, default='monai', required=False, help="User ID for the particle picks.")

    if add_slurm:
        slurm_group = parser.add_argument_group("SLURM Arguments")
        common.add_slurm_parameters(slurm_group, 'localize', gpus = 0)

    args = parser.parse_args()
    return args

# Entry point with argparse
def cli():
    
    parser_description = "Localized particles in tomograms using multiprocessing."
    args = localize_parser(parser_description)

    # Save JSON with Parameters
    output_yaml = f'localize_{args.pick_user_id}_{args.pick_session_id}.yaml'    
    save_parameters(args, output_yaml)    

    # Set multiprocessing start method
    mp.set_start_method("spawn")
    
    pick_particles(
        copick_config_path=args.config,
        method=args.method,
        seg_info=args.seg_info,
        voxel_size=args.voxel_size,
        pick_session_id=args.pick_session_id,
        pick_user_id=args.pick_user_id,
        radius_min_scale=args.radius_min_scale,
        radius_max_scale=args.radius_max_scale,
        filter_size=args.filter_size,
        runIDs=args.runIDs,
        pick_objects=args.pick_objects,
        n_procs=args.n_procs,
    )

def save_parameters(args: argparse.Namespace, 
                    output_path: str):

    # Organize parameters into categories
    params = {
        "input": {
            "config": args.config,
            "seg_name": args.seg_info[0],
            "seg_user_id": args.seg_info[1],
            "seg_session_id": args.seg_info[2],
            "voxel_size": args.voxel_size
        },
        "output": {
            "pick_session_id": args.pick_session_id,
            "pick_user_id": args.pick_user_id
        },
        "parameters": {
            "method": args.method,
            "radius_min_scale": args.radius_min_scale,
            "radius_max_scale": args.radius_max_scale,
            "filter_size": args.filter_size,
            "runIDs": args.runIDs
        }
    }

    # Print the parameters
    print(f"\nParameters for Localization:")
    pprint.pprint(params); print()

    # Save to YAML file
    utils.save_parameters_yaml(params, output_path)

if __name__ == "__main__":
    cli()

# def time_pick_particles():
#     import json, time

#     # Set multiprocessing start method
#     mp.set_start_method("spawn")

#     copick_config_path = "/mnt/simulations/ml_challenge/ml_config.json"  # Replace with your actual path
#     n_procs_list = [1, 4, 8, 16, 32]  # Adjust based on your needs
#     n_procs_list = [32, 16, 8, 4, 1]
#     timing_results = {}

#     session_id = 1
#     for n_procs in n_procs_list:
#         print(f"Testing with {n_procs} processes...")
#         start_time = time.time()
#         pick_particles(
#             copick_config_path=copick_config_path,
#             pick_session_id=str(session_id),
#             n_procs=n_procs
#         )
#         elapsed_time = time.time() - start_time
#         timing_results[n_procs] = elapsed_time
#         print(f"Elapsed time with {n_procs} processes: {elapsed_time:.2f} seconds")

#         session_id +=1 

#     # Save timing results to a JSON file
#     with open("timing_results.json", "w") as f:
#         json.dump(timing_results, f, indent=4)

#     print("Timing results saved to 'timing_results.json'")

# if __name__ == "__main__":
#     time_pick_particles()


from monai.inferers import sliding_window_inference
from monai.data import decollate_batch
from torch.multiprocessing import Pool
from monai.data import MetaTensor
from monai.transforms import (
    Compose, AsDiscrete, Activations
)
import octopi.processing.writers as write
from octopi.models import common
from typing import List, Optional
import torch, copick, gc, os
from octopi import io, utils
from tqdm import tqdm
import numpy as np

class Predictor:

    def __init__(self, 
                 config: str,
                 model_config: str,
                 model_weights: str,
                 apply_tta: bool = True,
                 device: Optional[str] = None):

        self.config = config
        self.root = copick.from_file(config)

        # Load the model config
        model_config = utils.load_yaml(model_config)

        self.Nclass = model_config['model']['num_classes']     
        self.dim_in = model_config['model']['dim_in']
        self.input_dim = None
        
        # Get the number of GPUs available
        num_gpus = torch.cuda.device_count()
        if num_gpus == 0:
            raise RuntimeError("No GPUs available.")

        # Set the device
        if device is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = device
        print('Running Inference On: ', self.device)

        # Check to see if the model weights file exists
        if not os.path.exists(model_weights):
            raise ValueError(f"Model weights file does not exist: {model_weights}")

        # Load the model weights
        model_builder = common.get_model(model_config['model']['architecture'])
        model_builder.build_model(model_config['model'])
        self.model = model_builder.model
        state_dict = torch.load(model_weights, map_location=self.device, weights_only=True)
        self.model.load_state_dict(state_dict)
        self.model.to(self.device)
        self.model.eval()

       # Initialize TTA if enabled
        self.apply_tta = apply_tta
        if self.apply_tta: 
            self.create_tta_augmentations() 
            # self.post_transforms = Compose([
            #     Activations(softmax=True)  # Keep probability output
            # ])
            self.softmax_transform = Compose([
                Activations(softmax=True)  # Keep probability output
            ])
            
            # Create the final discretization transform
            self.discretize_transform = AsDiscrete(argmax=True)
        else:
            # Define the post-processing transforms
            self.post_transforms = Compose([
                Activations(softmax=True),
                AsDiscrete(argmax=True)
            ])
            
        
    def _run_inference(self, input):
        """Apply sliding window inference to the input."""
        with torch.no_grad():
            predictions = sliding_window_inference(
                inputs=input,
                roi_size=(self.dim_in, self.dim_in, self.dim_in),
                sw_batch_size=4,  # one window is proecessed at a time
                predictor=self.model,
                overlap=0.5,
            )
            return [self.post_transforms(i) for i in decollate_batch(predictions)]

    def _run_inference_tta(self, input_data):
        """Memory-efficient TTA implementation that returns proper discrete segmentation maps."""
        
        batch_size = input_data.shape[0]
        results = []
        
        # Process one sample at a time
        for sample_idx in range(batch_size):
            # Extract single sample
            single_sample = input_data[sample_idx:sample_idx+1]
            
            # Initialize probability accumulator for this sample
            # Shape: [1, Nclass, Z, Y, X]
            acc_probs = torch.zeros(
                (1, self.Nclass, *single_sample.shape[2:]), 
                dtype=torch.float32, device=self.device
            )
            
            # Process each augmentation
            with torch.no_grad():
                for tta_transform, inverse_transform in zip(self.tta_transforms, self.inverse_tta_transforms):
                    # Apply transform to single sample
                    aug_sample = tta_transform(single_sample)
                    
                    # Free memory
                    torch.cuda.empty_cache()
                    
                    # Run inference (one sample at a time)
                    predictions = sliding_window_inference(
                        inputs=aug_sample,
                        roi_size=(self.dim_in, self.dim_in, self.dim_in),
                        sw_batch_size=4,  # Process one window at a time
                        predictor=self.model,
                        overlap=0.5,
                    )
                    
                    # Get softmax probabilities
                    probs = self.softmax_transform(predictions[0])  # Get first (only) item
                    
                    # Apply inverse transform with correct dimensions
                    inv_probs = inverse_transform(probs)
                    
                    # Accumulate probabilities
                    acc_probs[0] += inv_probs
                    
                    # Clear memory
                    del predictions, probs, inv_probs, aug_sample
            
            # Average accumulated probabilities
            acc_probs = acc_probs / len(self.tta_transforms)
            
            # Convert to discrete prediction - get argmax along class dimension
            # This gives us a tensor of shape [1, Z, Y, X] with discrete class indices
            discrete_pred = torch.argmax(acc_probs, dim=1)
            
            # Add to results - keeping only the spatial dimensions [Z, Y, X]
            results.append(discrete_pred[0])
            
            # Clear memory
            del acc_probs, discrete_pred
            torch.cuda.empty_cache()
        
        return results
        
    def predict_on_gpu(self, 
                        runIDs: List[str],
                        voxel_spacing: float,
                        tomo_algorithm: str ):

        # Load data for the current batch
        test_loader, test_dataset = io.create_predict_dataloader(
            self.root,
            voxel_spacing, tomo_algorithm,
            runIDs)
        
        # Determine Input Crop Size.
        if self.input_dim is None:
            self.input_dim = io.get_input_dimensions(test_dataset, self.dim_in)
        
        predictions = []
        with torch.no_grad():
            for data in tqdm(test_loader):
                tomogram = data['image'].to(self.device)
                if self.apply_tta: data['pred'] = self._run_inference_tta(tomogram)
                else:              data['pred']  = self._run_inference(tomogram)
                for idx in range(len(data['image'])):
                    predictions.append(data['pred'][idx].squeeze(0).numpy(force=True)) 

        return predictions

    def batch_predict(self, 
                      num_tomos_per_batch = 15, 
                      runIDs: Optional[str] = None,
                      voxel_spacing: float = 10,
                      tomo_algorithm: str = 'denoised', 
                      segmentation_name: str = 'prediction',
                      segmentation_user_id: str = 'octopi',
                      segmentation_session_id: str = '0'):

        """Run inference on tomograms in batches."""                          
        
        # If runIDs are not provided, load all runs
        if runIDs is None:
            runIDs = [run.name for run in self.root.runs if run.get_voxel_spacing(voxel_spacing) is not None]
            skippedRunIDs = [run.name for run in self.root.runs if run.get_voxel_spacing(voxel_spacing) is None]

            if skippedRunIDs:
                print(f"Warning: skipping runs with no voxel spacing {voxel_spacing}: {skippedRunIDs}")

        # Iterate over batches of runIDs
        for i in range(0, len(runIDs), num_tomos_per_batch):

            # Get a batch of runIDs
            batch_ids = runIDs[i:i + num_tomos_per_batch]  
            print('Running Inference on the Follow RunIDs: ', batch_ids)

            predictions = self.predict_on_gpu(batch_ids, voxel_spacing, tomo_algorithm)

            # Save Predictions to Corresponding RunID
            for ind in range(len(batch_ids)):
                run = self.root.get_run(batch_ids[ind])
                seg = predictions[ind]
                write.segmentation(run, seg, segmentation_user_id, segmentation_name, 
                                   segmentation_session_id, voxel_spacing)

            # After processing and saving predictions for a batch:
            del predictions  # Remove reference to the list holding prediction arrays
            torch.cuda.empty_cache()  # Clear unused GPU memory
            gc.collect()  # Trigger garbage collection for CPU memory

        print('Predictions Complete!')

    def create_tta_augmentations(self):
        """Define TTA augmentations and inverse transforms."""

        # Instead of Flip lets rotate around the first axis 3 times (90,180,270)
        self.tta_transforms = [
            lambda x: x,                    # Identity (no augmentation)
            lambda x: torch.rot90(x, k=1, dims=(3, 4)),  # 90° rotation
            lambda x: torch.rot90(x, k=2, dims=(3, 4)),  # 180° rotation
            lambda x: torch.rot90(x, k=3, dims=(3, 4)),  # 270° rotation
            # lambda x: torch.flip(x, dims=(3,)),        # Flip along height (spatial_axis=1)
            # lambda x: torch.flip(x, dims=(4,)),        # Flip along width (spatial_axis=2)  
            # lambda x: torch.flip(x, dims=(3, 4)),      # Flip along both height and width
        ]

        # Define inverse transformations (flip back to original orientation)
        self.inverse_tta_transforms = [
            lambda x: x,                           # Identity (no transformation needed)
            lambda x: torch.rot90(x, k=-1, dims=(2, 3)),  # Inverse of 90° (i.e. -90°)
            lambda x: torch.rot90(x, k=-2, dims=(2, 3)),  # Inverse of 180° (i.e. -180°)
            lambda x: torch.rot90(x, k=-3, dims=(2, 3)),  # Inverse of 270° (i.e. -270°)
            # lambda x: torch.flip(x, dims=(2,)),        # Same as forward
            # lambda x: torch.flip(x, dims=(3,)),        # Same as forward
            # lambda x: torch.flip(x, dims=(2, 3)),      # Same as forward
        ]

###################################################################################################################################################

class MultiGPUPredictor(Predictor):

    def __init__(self, 
                 config: str,
                 model_config: str,
                 model_weights: str):
        super().__init__(config, model_config, model_weights)
        self.num_gpus = torch.cuda.device_count()
        if self.num_gpus < 2:
            raise RuntimeError("MultiGPUPredictor requires at least 2 GPUs.")
    
    def predict_on_gpu(self, gpu_id: int, batch_ids: List[str], voxel_spacing: float, tomo_algorithm: str) -> List[np.ndarray]:
        """Helper function to run inference on a single GPU."""
        device = torch.device(f'cuda:{gpu_id}')
        self.model.to(device)
        
        # Load data specific to the batch assigned to this GPU
        test_loader = io.load_predict_data(self.root, batch_ids, voxel_spacing, tomo_algorithm)
        predictions = []
        
        with torch.no_grad():
            for data in tqdm(test_loader, desc=f"GPU {gpu_id}"):
                tomogram = data['image'].to(device)
                data["prediction"] = self.run_inference(tomogram)
                data = [self.post_processing(i) for i in decollate_batch(data)]
                for b in data:
                    predictions.append(b['prediction'].squeeze(0).cpu().numpy())
                    
        return predictions

    def multi_gpu_inference(self, 
                            num_tomos_per_batch: int = 15, 
                            runIDs: Optional[List[str]] = None,
                            voxel_spacing: float = 10,
                            tomo_algorithm: str = 'denoised', 
                            save: bool = False,
                            segmentation_name: str = 'prediction',
                            segmentation_user_id: str = 'monai',
                            segmentation_session_id: str = '0') -> Optional[List[np.ndarray]]:
        """Run inference across multiple GPUs, optionally saving results or returning predictions."""
        
        runIDs = runIDs or [run.name for run in self.root.runs]
        all_predictions = []

        # Divide runIDs into batches for each GPU
        batches = [runIDs[i:i + num_tomos_per_batch] for i in range(0, len(run_ids), num_tomos_per_batch)]
        
        # Run inference in parallel across GPUs
        for i in range(0, len(batches), self.num_gpus):
            gpu_batches = batches[i:i + self.num_gpus]
            with Pool(processes=self.num_gpus) as pool:
                results = pool.starmap(
                    self.predict_on_gpu,
                    [(gpu_id, gpu_batches[gpu_id], voxel_spacing, tomo_algorithm) for gpu_id in range(len(gpu_batches))]
                )

            # Collect and save results
            for gpu_id, predictions in enumerate(results):
                if save:
                    for idx, run_id in enumerate(gpu_batches[gpu_id]):
                        run = self.root.get_run(run_id)
                        segmentation = predictions[idx]
                        write.segmentation(run, segmentation, segmentation_user_id, segmentation_name, 
                                           segmentation_session_id, voxel_spacing)
                else:
                    all_predictions.extend(predictions)

        print('Multi-GPU predictions complete.')
        
        return None if save else all_predictions


from scipy.spatial.transform import Rotation as R
from octopi import utils, io
import scipy.ndimage as ndi
from typing import Tuple
import numpy as np
import math

def process_membrane_bound_extract(run,
                                   voxel_size: float,
                                   picks_info: Tuple[str, str, str],
                                   membrane_info: Tuple[str, str, str],
                                   organelle_info: Tuple[str, str, str],
                                   save_user_id: str,
                                   save_session_id: str,
                                   distance_threshold: float):

    """
    Process membrane-bound particles and extract their coordinates and orientations.
    
    Args:
        run: CoPick run object.
        voxel_size: Voxel size for coordinate scaling.
        segmentation_name: Name of the segmentation object.
        segmentation_user_id: User ID for the segmentation.
        segmentation_session_id: Session ID for the segmentation.
        picks_name: Name of the particle picks object.
        picks_user_id: User ID for the particle picks.
        picks_session_id: Session ID for the particle picks.
        save_user_id: User ID for saving processed picks.
        save_session_id: Session ID for saving close picks.
        distance_threshold: Maximum distance to consider a particle close to the membrane.
        organelle_seg: Whether to compute organelle centers from segmentation.
    """  

    # Increment session ID for the second class
    new_session_id = str(int(save_session_id) + 1)  # Convert to string after increment                                  

    # Need Better Error Handing for Missing Picks
    coordinates = io.get_copick_coordinates(
        run, 
        picks_info[0], picks_info[1], picks_info[2],
        voxel_size,
        raise_error=False
    )

    # If No Coordinates are Found, Return
    if coordinates is None:
        print(f'[Warning] RunID: {run.name} - No Coordinates Found for {picks_info[0]}, {picks_info[1]}, {picks_info[2]}')
        return

    nPoints = len(coordinates)

    # Determine which Segmentation to Use for Filtering
    if membrane_info is None:
        # Flag to distinguish between organelle and membrane segmentation
        membranes_provided = False
        seg = io.get_segmentation_array(run, 
                                    voxel_size, 
                                    organelle_info[0],
                                    user_id=organelle_info[1], 
                                    session_id=organelle_info[2],
                                    raise_error=False)
        # If No Segmentation is Found, Return
        if seg is None: return     
        elif nPoints == 0 or np.unique(seg).max() == 0:
            print(f'[Warning] RunID: {run.name} - Organelle-Seg Unique Values: {np.unique(seg)}, nPoints: {nPoints}')
            return                                            
    else:
        # Read both Organelle and Membrane Segmentations
        membranes_provided = True
        seg = io.get_segmentation_array(
            run, 
            voxel_size, 
            membrane_info[0],
            user_id=membrane_info[1], 
            session_id=membrane_info[2],
            raise_error=False)

        organelle_seg = io.get_segmentation_array(
            run, 
            voxel_size, 
            organelle_info[0],
            user_id=organelle_info[1], 
            session_id=organelle_info[2],
            raise_error=False)
        
        # If No Segmentation is Found, Return
        if seg is None or seg is None: return
        elif nPoints == 0 or np.unique(seg).max() == 0:
            print(f'[Warning] RunID: {run.name} - Organelle-Seg Unique Values: {np.unique(seg)}, nPoints: {nPoints}')
            return
        
        # Tempory Solution to Ensure Labels are the Same:
        seg[seg > 0] += 1

    if nPoints > 0:

        # Step 1: Find Closest Points to Segmentation of Interest
        points, closest_labels = closest_organelle_points(
            organelle_seg, 
            coordinates, 
            max_distance=distance_threshold, 
            return_labels_array=True
        )

        # Identify close and far indices
        close_indices = np.where(closest_labels != -1)[0]
        far_indices = np.where(closest_labels == -1)[0]

        # Initialize orientations array
        orientations = np.zeros([nPoints, 4, 4])
        orientations[:,3,3] = 1 

        # Step 2: Get Organelle Centers (Optional if an organelle segmentation is provided)
        organelle_centers = organelle_points(organelle_seg)

        # Step 3: Get All the Rotation Matrices from Euler Angles Based on Normal Vector
        if len(close_indices) > 0:

            # Get Organelle Centers for Close Points
            close_labels = closest_labels[close_indices]
            close_centers = np.array([organelle_centers[str(int(label))] for label in close_labels])

            # Calculate orientations
            for i, idx in enumerate(close_indices):
                rot = mCalcAngles(coordinates[idx], close_centers[i])
                r = R.from_euler('ZYZ', rot, degrees=True)
                orientations[idx,:3,:3] = r.inv().as_matrix()  

        # Swap z and x coordinates (0 and 2) before scaling Back to Angstroms
        coordinates[:, [0, 2]] = coordinates[:, [2, 0]]
        coordinates = coordinates * voxel_size
        
        # Save the close points in CoPick project
        if len(close_indices) > 0:
            try:
                close_picks = run.new_picks(object_name=picks_info[0], user_id=save_user_id, session_id=save_session_id)
            except:
                close_picks = run.get_picks(object_name=picks_info[0], user_id=save_user_id, session_id=save_session_id)[0]
            close_picks.from_numpy(coordinates[close_indices], orientations[close_indices])

        # Save the far points Coordinates in another CoPick pick
        if len(far_indices) > 0:                       
            try:
                far_picks = run.new_picks(object_name=picks_info[0], user_id=save_user_id, session_id=new_session_id)
            except:
                far_picks = run.get_picks(object_name=picks_info[0], user_id=save_user_id, session_id=new_session_id)[0]

            # Assume We Don't Know The Orientation for Anything Far From Membranes
            empty_orientations =  np.zeros(orientations[far_indices].shape)
            empty_orientations[:,-1,-1] = 1
            far_picks.from_numpy(coordinates[far_indices], empty_orientations)


def organelle_points(mask, xyz_order=False): 

    unique_labels = np.unique(mask)
    unique_labels = unique_labels[unique_labels > 0]  # Ignore background (label 0)

    coordinates = {}
    for label in unique_labels:
        center_of_mass = ndi.center_of_mass(mask == label)
        if xyz_order:
            center_of_mass = center_of_mass[::-1]
        coordinates[str(label)] = center_of_mass
        # coordinates[str(label)] = ndimage.center_of_mass(mask == label)
    return coordinates     

def closest_organelle_points(mask, coords, min_distance = 0, max_distance=float('inf'), return_labels_array=False):
    """
    Filter points in `coords` based on their proximity to the lysosome membrane.

    Args:
        mask (numpy.ndarray): 3D segmentation mask with integer labels.
        coords (numpy.ndarray): Array of shape (N, 3) with 3D coordinates.
        min_distance (float): Minimum distance threshold for a point to be considered.
        max_distance (float): Maximum distance threshold for a point to be considered.
        return_labels_array (bool): Whether to return the labels array matching the
                                    original order of coords.

    Returns:
        dict: A dictionary where keys are mask labels and values are lists of points
              (3D coordinates) within the specified distance range.
        numpy.ndarray (optional): Array of shape (N,) with the label for each coordinate,
                                   or -1 if the point is outside the specified range.
                                   Only returned if `return_labels_array=True`.
    """

    unique_labels = np.unique(mask)
    unique_labels = unique_labels[unique_labels > 0]  # Ignore background (label 0)

    # Combine all mask points and keep track of their labels
    all_mask_points = []
    all_labels = []
    for label in unique_labels:
        label_points = np.argwhere(mask == label)
        all_mask_points.append(label_points)
        all_labels.extend([label] * len(label_points))

    # Combine all mask points and labels into arrays
    all_mask_points = np.vstack(all_mask_points)
    all_labels = np.array(all_labels)    

    # Initialize a dictionary to store filtered points for each label
    label_to_filtered_points = {label: [] for label in unique_labels}
    label_to_filtered_points['far'] = []  # Initialize 'far' key to store rejected points    

    # Initialize an array to store the closest label or -1 for out-of-range points
    closest_labels = np.full(len(coords), -1, dtype=int)

    # Compute the closest label and filter based on distance
    for i, coord in enumerate(coords):
        distances = np.linalg.norm(all_mask_points - coord, axis=1)
        min_index = np.argmin(distances)
        closest_label = all_labels[min_index]
        min_distance_to_membrane = distances[min_index]

        # Check if the distance is within the allowed range
        if min_distance <= min_distance_to_membrane <= max_distance:
            closest_labels[i] = closest_label
            label_to_filtered_points[closest_label].append(coord)
        else:
            label_to_filtered_points['far'].append(coord)

    # Convert lists to NumPy arrays for easier handling
    for label in label_to_filtered_points:
        label_to_filtered_points[label] = np.array(label_to_filtered_points[label])

    if return_labels_array:
        return label_to_filtered_points, closest_labels
    else:
        # Concatenate all points into a single NumPy array
        concatenated_points = np.vstack([points for points in label_to_filtered_points.values() if points.size > 0])        
        return concatenated_points

# Create Class to Estimate Eulers from Centers of Lysate
def mCalcAngles(mbProtein, membrane_point):

    deltaX = mbProtein[0] - membrane_point[0]
    deltaY = mbProtein[1] - membrane_point[1]
    deltaZ = mbProtein[2] - membrane_point[2]
    #-----------------------------
    # angRotion is in [-180, 180]
    #-----------------------------
    angRot = math.atan(deltaY / (deltaX + 1e-30))
    angRot *= (180 / math.pi)
    if deltaX < 0 and deltaY > 0:
        angRot += 180
    elif deltaX < 0 and deltaY < 0:
        angRot -= 180
    angRot = float("{:.2f}".format(angRot))
    #------------------------
    # angTilt is in [0, 180]
    #------------------------
    rXY = math.sqrt(deltaX * deltaX + deltaY * deltaY)
    angTilt = math.atan(rXY / (deltaZ + 1e-30))
    angTilt *= (180 / math.pi)
    if angTilt < 0:
        angTilt += 180.0
    angTilt = float("{:.2f}".format(angTilt))

    return (angRot, angTilt, 0) 


from typing import List, Tuple, Callable, Optional, Dict, Any
from monai.transforms import Compose
from monai.data import CacheDataset
from octopi import io
from tqdm import tqdm
import os, sys

class MultiConfigCacheDataset(CacheDataset):
    """
    A custom CacheDataset that loads data lazily from multiple sources
    with consolidated loading and caching process.
    """
    
    def __init__(
        self,
        manager,
        run_ids: List[Tuple[str, str]],
        transform: Optional[Callable] = None,
        cache_rate: float = 1.0,
        num_workers: int = 0,
        progress: bool = True,
        copy_cache: bool = True,
        cache_num: int = sys.maxsize
    ):
        # Save reference to manager and run_ids
        self.manager = manager
        self.run_ids = run_ids
        self.progress = progress
        
        # Prepare empty data list first - don't load immediately
        self.data = []
        
        # Initialize the parent CacheDataset with an empty list
        # We'll override the _fill_cache method to handle loading and caching in one step
        super().__init__(
            data=[],  # Empty list - we'll load data in _fill_cache
            transform=transform,
            cache_rate=cache_rate,
            num_workers=num_workers,
            progress=False,  # We'll handle our own progress
            copy_cache=copy_cache,
            cache_num=cache_num
        )
    
    def _fill_cache(self):
        """
        Override the parent's _fill_cache method to combine loading and caching.
        """
        if self.progress:
            print("Loading and caching dataset...")
        
        # Load and process data in a single operation
        self.data = []
        iterator = tqdm(self.run_ids, desc="Loading dataset") if self.progress else self.run_ids
        
        for session_name, run_name in iterator:
            root = self.manager.roots[session_name]
            batch_data = io.load_training_data(
                root, 
                [run_name], 
                self.manager.voxel_size, 
                self.manager.tomo_algorithm, 
                self.manager.target_name, 
                self.manager.target_session_id, 
                self.manager.target_user_id, 
                progress_update=False
            )
            
            self.data.extend(batch_data)
            
            # Process and cache this batch right away
            for i, item in enumerate(batch_data):
                if len(self._cache) < self.cache_num and self.cache_rate > 0.0:
                    if np.random.random() < self.cache_rate:
                        self._cache.append(self._transform(item))
        
        # Check max label value if needed
        if hasattr(self.manager, '_check_max_label_value'):
            self.manager._check_max_label_value(self.data)
        
        # Update the _data attribute to match the loaded data
        self._data = self.data
    
    def __len__(self):
        """
        Return the length of the dataset.
        """
        if not self.data:
            self._fill_cache()  # Load data if not loaded yet
        return len(self.data)
    
    def __getitem__(self, index):
        """
        Return the item at the given index.
        """
        if not self.data:
            self._fill_cache()  # Load data if not loaded yet
        
        # Use parent's logic for cached items
        if index < len(self._cache):
            return self._cache[index]
        
        # Otherwise transform on-the-fly
        return self._transform(self.data[index])

# TODO: Implement Single Config Cache Dataset
# class SingleConfigCacheDataset(CacheDataset):
#     def __init__(self, 
#                  root: Any, 
#                  run_ids: List[str], 
#                  voxel_size: float, 
#                  tomo_algorithm: str, 
#                  target_name: str, 


from octopi import utils, io
from scipy.spatial import distance
from typing import List
import copick, json, os
import numpy as np

class evaluator:

    def __init__(self, 
                 copick_config: str,
                 ground_truth_user_id: str,
                 ground_truth_session_id: str,
                 prediction_user_id: str,
                 predict_session_id: str,
                 voxel_size: float = 10,
                 beta: float = 4,
                 object_names: List[str] = None):
        
        self.root = copick.from_file(copick_config)
        print('Running Evaluation on the Following Copick Project: ', copick_config)

        self.ground_truth_user_id = ground_truth_user_id
        self.ground_truth_session_id = ground_truth_session_id
        self.prediction_user_id = prediction_user_id
        self.predict_session_id = predict_session_id
        self.voxel_size = voxel_size
        self.beta = beta
        print(f'\nGround Truth Query: \nUserID: {ground_truth_user_id}, SessionID: {ground_truth_session_id}')
        print(f'\nSubmitted Picks: \nUserID: {prediction_user_id}, SessionID: {predict_session_id}\n')

        # Save input parameters
        self.input_params = {
            "copick_config": copick_config,
            "ground_truth_user_id": ground_truth_user_id,
            "ground_truth_session_id": ground_truth_session_id,
            "prediction_user_id": prediction_user_id,
            "predict_session_id": predict_session_id,
        }

        # Get objects that can be Picked
        if not object_names:
            print('No object names provided, using all pickable objects')
            self.objects = [(obj.name, obj.radius) for obj in self.root.pickable_objects if obj.is_particle]
        else:
            # Get valid pickable objects with their radii
            valid_objects = {obj.name: obj.radius for obj in self.root.pickable_objects if obj.is_particle}
            
            # Filter and validate provided object names
            invalid_objects = [name for name in object_names if name not in valid_objects]
            if invalid_objects:
                print('WARNING: The following object names are not valid pickable objects:', invalid_objects)
                print('Valid objects are:', list(valid_objects.keys()))
            
            self.objects = [(name, valid_objects[name]) for name in object_names if name in valid_objects]
            
            if not self.objects:
                raise ValueError("None of the provided object names are valid pickable objects")
                
        print('Using the following valid objects:', [name for name, _ in self.objects])

        # Define object-specific weights
        self.weights = {
            "apo-ferritin": 1,
            "beta-amylase": 0,  # Excluded from scoring
            "beta-galactosidase": 2,
            "ribosome": 1,
            "thyroglobulin": 2,
            "virus-like particle": 1,
        }

    def run(self, 
            save_path: str = None,
            distance_threshold_scale: float = 0.8,
            runIDs: List[str] = None):
    
        # Type check for runIDs
        if runIDs is not None and not (isinstance(runIDs, list) and all(isinstance(x, str) for x in runIDs)):
            raise TypeError("runIDs must be a list of strings")

        run_ids = runIDs if runIDs else [run.name for run in self.root.runs]
        print('\nRunning Metrics Evaluation on the Following RunIDs: ', run_ids)

        metrics = {}
        summary_metrics = {name: {'precision': [], 'recall': [], 'f1_score': [], 'fbeta_score': [], 'accuracy': [], 
                                'true_positives': [], 'false_positives': [], 'false_negatives': []} for name, _ in self.objects}
        
        # For storing the aggregated counts per particle type (across all runs)
        aggregated_counts = {name: {'total_tp': 0, 'total_fp': 0, 'total_fn': 0} for name, _ in self.objects}

        for runID in run_ids:
            # Initialize the nested dictionary for this runID
            metrics[runID] = {}
            run = self.root.get_run(runID)

            for name, radius in self.objects:
                
                # Get Ground Truth and Predicted Coordinates
                gt_coordinates = io.get_copick_coordinates(
                    run, name, 
                    self.ground_truth_user_id, self.ground_truth_session_id, 
                    self.voxel_size, raise_error=False
                )
                pred_coordinates = io.get_copick_coordinates(
                    run, name,
                    self.prediction_user_id, self.predict_session_id, 
                    self.voxel_size, raise_error=False
                )
                
                # If no reference (GT) points, all candidate points are false positives
                if gt_coordinates is None or len(gt_coordinates) == 0:
                    num_pred_points = pred_coordinates.shape[0] if pred_coordinates is not None else 0
                    metrics[runID][name] = {'precision': 0, 'recall': 0, 'fbeta_score': 0, 'true_positives': 0, 'false_positives': num_pred_points, 'false_negatives': 0}
                    
                    # Update aggregated counts
                    aggregated_counts[name]['total_fp'] += num_pred_points
                    
                    continue

                # If no candidate (predicted) points, all reference points are false negatives
                if pred_coordinates is None or len(pred_coordinates) == 0:
                    num_gt_points = gt_coordinates.shape[0] if gt_coordinates is not None else 0
                    metrics[runID][name] = {'precision': 0, 'recall': 0, 'fbeta_score': 0, 'true_positives': 0, 'false_positives': 0, 'false_negatives': num_gt_points}
                    
                    # Update aggregated counts
                    aggregated_counts[name]['total_fn'] += num_gt_points
                    
                    continue

                # Compute Distance Threshold Based on Particle Radius
                distance_threshold = (radius/self.voxel_size) * distance_threshold_scale
                metrics[runID][name] = self.compute_metrics(gt_coordinates, pred_coordinates, distance_threshold)         

                # Collect metrics for summary statistics
                for key in summary_metrics[name]:
                    summary_metrics[name][key].append(metrics[runID][name][key])
                
                # Update aggregated counts
                aggregated_counts[name]['total_tp'] += metrics[runID][name]['true_positives']
                aggregated_counts[name]['total_fp'] += metrics[runID][name]['false_positives']
                aggregated_counts[name]['total_fn'] += metrics[runID][name]['false_negatives']

        # Create a new dictionary for summarized metrics
        final_summary_metrics = {}

        # Compute average metrics and standard deviations across runs for each object
        for name, _ in self.objects:
            # Initialize the final summary for the object
            final_summary_metrics[name] = {}

            for key in summary_metrics[name]:
                mu_val = float(np.mean(summary_metrics[name][key]))
                std_val = float(np.std(summary_metrics[name][key]))

                # Populate the new dictionary with structured data
                final_summary_metrics[name][key] = {
                    'mean': mu_val,
                    'std': std_val
                }

        print('\nAverage Metrics Summary:')
        self.print_metrics_summary(final_summary_metrics)          
        
        # Compute Final Kaggle Submission Score using reference approach
        aggregate_fbeta = 0.0
        total_weight = 0.0
        
        print('\nCalculating Final F-beta Score using per-particle approach:')
        for name, counts in aggregated_counts.items():
            tp = counts['total_tp']
            fp = counts['total_fp']
            fn = counts['total_fn']
            
            # Calculate precision and recall for this particle type
            precision = tp / (tp + fp) if (tp + fp) > 0 else 0
            recall = tp / (tp + fn) if (tp + fn) > 0 else 0
            
            # Calculate F-beta for this particle type
            particle_fbeta = (1 + self.beta**2) * (precision * recall) / \
                            ((self.beta**2 * precision) + recall) if \
                            ((self.beta**2 * precision) + recall) > 0 else 0
            
            # Get the weight for this particle type
            weight = self.weights.get(name, 1)
            
            # Accumulate weighted F-beta score
            aggregate_fbeta += particle_fbeta * weight
            total_weight += weight
            
            print(f"  {name}: TP={tp}, FP={fp}, FN={fn}, Precision={precision:.3f}, " + 
                f"Recall={recall:.3f}, F-beta={particle_fbeta:.3f}, Weight={weight}")
        
        # Normalize by total weight
        final_fbeta = aggregate_fbeta / total_weight if total_weight > 0 else 0
        
        print(f'\nFinal Kaggle Submission Score: {final_fbeta:.3f}')   

        # Save average and detailed metrics with parameters included        
        if save_path:
            self.parameters = {
                "distance_threshold_scale": distance_threshold_scale,
                "runIDs": runIDs,
            }    
            
            os.makedirs(save_path, exist_ok=True)
            summary_metrics = { "input": self.input_params, "parameters": self.parameters, 
                                    "summary_metrics": final_summary_metrics }
            with open(os.path.join(save_path, 'average_metrics.json'), 'w') as f:
                json.dump(summary_metrics, f, indent=4)
            print(f'\nAverage Metrics saved to {os.path.join(save_path, "average_metrics.json")}')
            
            detailed_metrics = { "input": self.input_params, "parameters": self.parameters, 
                                "metrics": metrics }
            with open(os.path.join(save_path, 'metrics.json'), 'w') as f:
                json.dump(detailed_metrics, f, indent=4)
            print(f'Metrics saved to {os.path.join(save_path, "metrics.json")}')

    def compute_metrics(self, 
                        gt_points, 
                        pred_points, 
                        threshold):
        
        gt_points = np.array(gt_points)
        pred_points = np.array(pred_points)
        
        # Calculate distances
        if gt_points.shape[0] == 0:
            # No ground truth points: all predictions are false positives
            fp = pred_points.shape[0]
            fn = 0
            tp = 0
        elif pred_points.shape[0] == 0:
            # No predictions: all ground truth points are false negatives
            fp = 0
            fn = gt_points.shape[0]
            tp = 0
        else:    
            # Calculate distances
            dist_matrix = distance.cdist(pred_points, gt_points, 'euclidean')

            # Determine matches within the threshold
            tp = np.sum(np.min(dist_matrix, axis=1) < threshold)
            fp = np.sum(np.min(dist_matrix, axis=1) >= threshold)
            fn = np.sum(np.min(dist_matrix, axis=0) >= threshold)
        
        # Precision, Recall, F1 Score
        precision = tp / (tp + fp) if tp + fp > 0 else 0
        recall = tp / (tp + fn) if tp + fn > 0 else 0
        f1_score = 2 * (precision * recall) / (precision + recall) if precision + recall > 0 else 0
        accuracy = tp / (tp + fp + fn)  # Note: TN not considered here

         # Compute F_beta using the formula
        if (self.beta**2 * precision + recall) > 0:
            fbeta = (1 + self.beta**2) * (precision * recall) / (self.beta**2 * precision + recall)
        else:
            fbeta = 0
        
        return {
            'precision': precision,
            'recall': recall,
            'f1_score': f1_score,
            'fbeta_score': fbeta,
            'accuracy': accuracy, 
            'true_positives': int(tp),
            'false_positives': int(fp),
            'false_negatives': int(fn)
        }
    
    def print_metrics_summary(self, metrics_dict):
        for name, metrics in metrics_dict.items():
            recall = metrics['recall']
            precision = metrics['precision']
            f1_score = metrics['f1_score']
            fbeta_score = metrics['fbeta_score']
            false_positives = metrics['false_positives']
            false_negatives = metrics['false_negatives']
            
            # Format the metrics for the current object
            formatted_metrics = (
                f"Recall: {recall['mean']:.3f} ± {recall['std']:.3f}, "
                f"Precision: {precision['mean']:.3f} ± {precision['std']:.3f}, "
                f"F1 Score: {f1_score['mean']:.3f} ± {f1_score['std']:.3f}, "
                f"F_beta Score: {fbeta_score['mean']:.3f} ± {fbeta_score['std']:.3f}, "
                f"False_Positives: {false_positives['mean']:.1f} ± {false_positives['std']:.1f}, "
                f"False_Negatives: {false_negatives['mean']:.1f} ± {false_negatives['std']:.1f}"
            )
            
            # Print the object name and its metrics
            print(f"{name}: [{formatted_metrics}]")
        print()


import octopi.processing.create_targets_from_picks as create_targets
from typing import List, Tuple, Union
from collections import defaultdict
import argparse, copick, yaml, os
from octopi import utils, io 
from tqdm import tqdm
import numpy as np

def create_sub_train_targets(
    config: str,
    pick_targets: List[Tuple[str, Union[str, None], Union[str, None]]],  # Updated type without radius
    seg_targets: List[Tuple[str, Union[str, None], Union[str, None]]],
    voxel_size: float,
    radius_scale: float,    
    tomogram_algorithm: str,
    target_segmentation_name: str,
    target_user_id: str,
    target_session_id: str,
    run_ids: List[str],    
    ):

    # Load Copick Project 
    root = copick.from_file(config)

    # Create empty dictionary for all targets
    train_targets = defaultdict(dict)

    # Create dictionary for particle targets
    for t in pick_targets:
        # Parse the target
        obj_name, user_id, session_id = t 
        obj = root.get_object(obj_name)

        # Check if the object is valid
        if obj is None:
            print(f'Warning - Skipping Particle Target: "{obj_name}", as it is not a valid name in the config file.')
            continue
        
        # Get the label and radius of the object
        label = obj.label
        info = {
            "label": label,
            "user_id": user_id,
            "session_id": session_id,
            "is_particle_target": True,
            "radius": root.get_object(obj_name).radius,
        }
        train_targets[obj_name] = info

    # Create dictionary for segmentation targets
    train_targets = add_segmentation_targets(root, seg_targets, train_targets)

    create_targets.generate_targets(
        root, train_targets, voxel_size, tomogram_algorithm, radius_scale,
        target_segmentation_name, target_user_id, 
        target_session_id, run_ids
    )


def create_all_train_targets(
    config: str,
    seg_targets: List[List[Tuple[str, Union[str, None], Union[str, None]]]],
    picks_session_id: str,
    picks_user_id: str,
    voxel_size: float,
    radius_scale: float,
    tomogram_algorithm: str,
    target_segmentation_name: str,
    target_user_id: str,
    target_session_id: str,
    run_ids: List[str],    
    ):     

    # Load Copick Project 
    root = copick.from_file(config)

    # Create empty dictionary for all targets
    target_objects = defaultdict(dict)

    # Create dictionary for particle targets
    for object in root.pickable_objects:
        info = {
            "label": object.label,
            "radius": object.radius,
            "user_id": picks_user_id,
            "session_id": picks_session_id,
            "is_particle_target": True,
        }
        target_objects[object.name] = info

    # Create dictionary for segmentation targets
    target_objects = add_segmentation_targets(root, seg_targets, target_objects)

    create_targets.generate_targets(
        root, target_objects, voxel_size, tomogram_algorithm, 
        radius_scale, target_segmentation_name, target_user_id, 
        target_session_id, run_ids 
    )

def add_segmentation_targets(
    root,
    seg_targets,
    train_targets: dict,
    ):

    # Create dictionary for segmentation targets
    for s in seg_targets:

        # Parse Segmentation Target
        obj_name, user_id, session_id = s

        # Add Segmentation Target
        try:
            info = {
                "label": root.get_object(obj_name).label,
                "user_id": user_id,
                "session_id": session_id,
                "is_particle_target": False,                 
                "radius": None,    
            }
            train_targets[obj_name] = info

        # If Segmentation Target is not found, print warning
        except:
            print(f'Warning - Skipping Segmentation Name: "{obj_name}", as it is not a valid object in the Copick project.')

    return train_targets    

def parse_args():
    """
    Parse command-line arguments for generating segmentation targets from CoPick configurations.

    This tool allows researchers to specify protein labels for training in two ways:
    
    1. **Manual Specification:** Users can define a subset of pickable objects from the CoPick configuration file.
       - Specify a target protein using `--target name`, or refine selection with `--target name,user_id,session_id`.
       - This enables flexible training target customization from multiple sources.

    2. **Automated Query:** Instead of specifying targets explicitly, users can provide a session ID (`--picks-session-id`) and/or 
       user ID (`--picks-user-id`). DeepFindET will automatically retrieve all pickable objects associated with the query.
    
    The tool also allows customization of tomogram reconstruction settings and segmentation parameters.

    Example Usage:
        - Manual Specification:
            ```bash
            python create_targets.py --config config.json --target ribosome --target apoferritin,123,456
            ```
        - Automated Query:
            ```bash
            python create_targets.py --config config.json --picks-session-id 123 --picks-user-id 456
            ```

    Output segmentation data is saved in a structured YAML file.
    """
    parser = argparse.ArgumentParser(
        description=f"""Generate segmentation targets from CoPick configurations with either --target flag (which lets users specify a subset of pickable objects) or --picks-session-id and --picks-user-id flags (which lets users specify a sessionID and userID to automatically retrieve all pickable objects associated with the query).""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    input_group = parser.add_argument_group("Input Arguments")
    input_group.add_argument("--config", type=str, required=True, help="Path to the CoPick configuration file.")
    input_group.add_argument("--target", type=utils.parse_target, action="append", default=None, help='Target specifications: "name" or "name,user_id,session_id".')
    input_group.add_argument("--picks-session-id", type=str, default=None, help="Session ID for the picks.")
    input_group.add_argument("--picks-user-id", type=str, default=None, help="User ID associated with the picks.")
    input_group.add_argument("--seg-target", type=utils.parse_target, action="append", default=[], help='Segmentation targets: "name" or "name,user_id,session_id".')
    input_group.add_argument("--run-ids", type=utils.parse_list, default=None, help="List of run IDs.")

    # Parameters
    parameters_group = parser.add_argument_group("Parameters")
    parameters_group.add_argument("--tomo-alg", type=str, default="wbp", help="Tomogram reconstruction algorithm.")    
    parameters_group.add_argument("--radius-scale", type=float, default=0.8, help="Scale factor for object radius.")
    parameters_group.add_argument("--voxel-size", type=float, default=10, help="Voxel size for tomogram reconstruction.")    
    
    output_group = parser.add_argument_group("Output Arguments")
    output_group.add_argument("--target-segmentation-name", type=str, default='targets', help="Name for the target segmentation.")
    output_group.add_argument("--target-user-id", type=str, default="octopi", help="User ID associated with the target segmentation.")
    output_group.add_argument("--target-session-id", type=str, default="1", help="Session ID for the target segmentation.")

    return parser.parse_args()

def cli():
    args = parse_args()

    # Save JSON with Parameters
    output_yaml = f'create-targets_{args.target_user_id}_{args.target_session_id}_{args.target_segmentation_name}.yaml'
    save_parameters(args, output_yaml)      

    if args.target:
        # If at least one --target is provided, call create_sub_train_targets
        create_sub_train_targets(
            config=args.config,
            pick_targets=args.target,
            seg_targets=args.seg_target,
            voxel_size=args.voxel_size,
            radius_scale=args.radius_scale,
            tomogram_algorithm=args.tomo_alg,
            target_segmentation_name=args.target_segmentation_name,
            target_user_id=args.target_user_id,
            target_session_id=args.target_session_id,
            run_ids=args.run_ids,
        )
    else:
        # If no --target is provided, call create_all_train_targets
        create_all_train_targets(
            config=args.config,
            seg_targets=args.seg_target,
            picks_session_id=args.picks_session_id,
            picks_user_id=args.picks_user_id,
            voxel_size=args.voxel_size,
            radius_scale=args.radius_scale,
            tomogram_algorithm=args.tomo_alg,
            target_segmentation_name=args.target_segmentation_name,
            target_user_id=args.target_user_id,
            target_session_id=args.target_session_id,
            run_ids=args.run_ids,
        )

def save_parameters(args, output_path: str):
    """
    Save parameters to a YAML file with subgroups for input, output, and parameters.
    Append to the file if it already exists.

    Args:
        args: Parsed arguments from argparse.
        output_path: Path to save the YAML file.
    """

    print('\nGenerating Target Segmentation Masks from the Following Copick-Query:')
    if args.picks_session_id is None or args.picks_user_id is None:
        print(f'    - {args.target}\n')
        input_group = {
            "config": args.config,
            "target": args.target,
        }
    else:
        print(f'    - {args.picks_session_id}, {args.picks_user_id}\n')
        input_group = {
            "config": args.config,
            "picks_session_id": args.picks_session_id,
            "picks_user_id": args.picks_user_id
        }
    if len(args.seg_target) > 0:
        input_group["seg_target"] = args.seg_target
        
    # Organize parameters into subgroups
    input_key = f'{args.target_user_id}_{args.target_session_id}_{args.target_segmentation_name}'
    new_entry = {
        input_key : {
            "input": input_group ,
            "parameters": {
                "radius_scale": args.radius_scale,
                "tomogram_algorithm": args.tomo_alg,
                "voxel_size": args.voxel_size,            
            }
        }
    }

     # Check if the YAML file already exists
    if os.path.exists(output_path):
        # Load the existing content
        with open(output_path, 'r') as f:
            try:
                existing_data = yaml.safe_load(f)
                if existing_data is None:
                    existing_data = {}  # Ensure it's a dictionary
                elif not isinstance(existing_data, dict):
                    raise ValueError("Existing YAML data is not a dictionary. Cannot update.")
            except yaml.YAMLError:
                existing_data = {}  # Treat as empty if the file is malformed
    else:
        existing_data = {}  # Initialize as empty list if the file does not exist

    # Append the new entry
    existing_data[input_key] = new_entry[input_key]

    # Save back to the YAML file
    utils.save_parameters_yaml(existing_data, output_path)

if __name__ == "__main__":
    cli()
